# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# At the time I created this plugin, their website had lots of issues with the
# Firefox Browser.  I fell back to Chrome and had no problems.

# There are 2 ways to use this plugin...
#
# Method 1: Via Webhook (default mode):
#   Visit https://teams.webex.com and make yourself an account if you don't
#   already have one.  You'll want to create at least one 'space' before
#   getting the 'incoming webhook'.
#
#   Next you'll need to install the 'Incoming webhook' plugin found under
#   the 'other' category here: https://apphub.webex.com/integrations/
#
#   These links may not always work as time goes by and websites always
#   change, but at the time of creating this plugin this was a direct link
#   to it:
#     https://apphub.webex.com/integrations/incoming-webhooks-cisco-systems
#
#   If you're logged in, you'll be able to click on the 'Connect' button.
#   From there you'll need to accept the permissions it will ask of you.
#   Give the webhook a name such as 'apprise'.
#   When you're complete, you will recieve a URL that looks something like:
#     https://api.ciscospark.com/v1/webhooks/incoming/\
#           Y3lzY29zcGkyazovL3VzL1dFQkhPT0sajkkzYWU4fTMtMGE4Yy00
#
#   The last part of the URL is all you need to be interested in. Think of
#   this url as:
#     https://api.ciscospark.com/v1/webhooks/incoming/{token}
#
#   You will need to assemble all of your URLs for this plugin to work as:
#     wxteams://{token}
#
#
# Method 2: Via Bot/API (supports file attachments):
#   Visit https://developer.webex.com/my-apps and create a new Bot.
#   After creating the bot, you'll receive a Bot Access Token.
#
#   You will also need to know the Room ID(s) you want to post to.
#   Room IDs can be retrieved from the Webex API:
#     https://developer.webex.com/docs/api/v1/rooms/list-rooms
#
#   Assemble your Apprise URL as:
#     wxteams://{access_token}/{room_id}
#     wxteams://{access_token}/{room_id1}/{room_id2}
#
#   The plugin auto-detects the mode:
#     - If the token is 80-160 lowercase alphanumeric chars -> Webhook mode
#     - Otherwise -> Bot mode (requires at least one room ID)
#   You may also force the mode with ?mode=webhook or ?mode=bot.
#
# Resources
# - https://developer.webex.com/docs/api/basics - markdown/post syntax
# - https://developer.webex.com/docs/api/v1/messages/create-a-message
# - https://developer.cisco.com/ecosystem/webex/apps/\
#       incoming-webhooks-cisco-systems/ - Simple webhook example

from json import dumps
import re

import requests

from ..common import NotifyFormat, NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import parse_list, validate_regex
from .base import NotifyBase

# Extend HTTP Error Messages
# Based on: https://developer.webex.com/docs/api/basics/rate-limiting
WEBEX_HTTP_ERROR_MAP = {
    401: "Unauthorized - Invalid Token.",
    415: "Unsuported media specified",
    429: "To many consecutive requests were made.",
    503: "Service is overloaded, try again later",
}


class WebexTeamsMode:
    """Tracks the mode of which we're using Webex Teams."""

    # We're dealing with an incoming webhook
    # Token is 80-160 lowercase alphanumeric chars
    WEBHOOK = "webhook"

    # We're dealing with a Bot/API access token (supports attachments)
    # Token is a Bearer access token from the Webex developer portal
    BOT = "bot"


# Define our Webex Teams Modes
WEBEX_TEAMS_MODES = (
    WebexTeamsMode.WEBHOOK,
    WebexTeamsMode.BOT,
)


class NotifyWebexTeams(NotifyBase):
    """A wrapper for Webex Teams Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Cisco Webex Teams"

    # The services URL
    service_url = "https://webex.teams.com/"

    # The default secure protocol
    secure_protocol = ("wxteams", "webex")

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://appriseit.com/services/wxteams/"

    # Webex Teams Webhook URL
    notify_url = "https://api.ciscospark.com/v1/webhooks/incoming/"

    # Webex Teams Bot API URL (used in Bot mode)
    api_url = "https://webexapis.com/v1/messages"

    # Bot mode supports attachments
    attachment_support = True

    # Do not set body_maxlen as it is set in a property value below
    # since the length varies depending on whether we are using a
    # webhook (1000 chars) or the bot API (7439 chars)
    # body_maxlen = see below @property defined

    # We don't support titles for Webex notifications
    title_maxlen = 0

    # Default to markdown; fall back to text
    notify_format = NotifyFormat.MARKDOWN

    # Define object URL templates
    templates = (
        # Webhook mode (existing)
        "{schema}://{token}",
        # Bot mode (access_token in host, one or more room IDs in path)
        "{schema}://{access_token}/{targets}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "token": {
                "name": _("Webhook Token"),
                "type": "string",
                "private": True,
                "required": True,
                "regex": (r"^[a-z0-9]{80,160}$", "i"),
            },
            "access_token": {
                "name": _("Bot Access Token"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "targets": {
                "name": _("Room IDs"),
                "type": "list:string",
            },
        },
    )

    template_args = dict(
        NotifyBase.template_args,
        **{
            "token": {
                "alias_of": "token",
            },
            "mode": {
                "name": _("Mode"),
                "type": "choice:string",
                "values": WEBEX_TEAMS_MODES,
                # mode is auto-detected if not specified
            },
            "to": {
                "alias_of": "targets",
            },
        },
    )

    def __init__(
        self,
        token=None,
        access_token=None,
        targets=None,
        mode=None,
        **kwargs,
    ):
        """Initialize Webex Teams Object."""
        super().__init__(**kwargs)

        # Resolve mode: explicit override wins, otherwise auto-detect
        if mode and isinstance(mode, str):
            self.mode = next(
                (m for m in WEBEX_TEAMS_MODES if m.startswith(mode)), None
            )
            if self.mode not in WEBEX_TEAMS_MODES:
                msg = f"The Webex Teams mode specified ({mode}) is invalid."
                self.logger.warning(msg)
                raise TypeError(msg)

        else:
            # Auto-detect: webhook tokens are 80-160 lowercase alphanumeric
            _candidate = access_token or token
            if token and validate_regex(
                token, *self.template_tokens["token"]["regex"]
            ):
                self.mode = WebexTeamsMode.WEBHOOK

            elif _candidate:
                # Non-webhook token length/format -> assume BOT
                self.mode = WebexTeamsMode.BOT

            else:
                # No usable token at all; will fail below
                self.mode = WebexTeamsMode.WEBHOOK

        if self.mode == WebexTeamsMode.WEBHOOK:
            self.access_token = None
            self.targets = []

            # Webhook token: prefer 'token', fall back to 'access_token'
            _tok = token or access_token
            self.token = validate_regex(
                _tok, *self.template_tokens["token"]["regex"]
            )
            if not self.token:
                msg = (
                    "The Webex Teams webhook token"
                    f" specified ({_tok}) is invalid."
                )
                self.logger.warning(msg)
                raise TypeError(msg)

        else:  # WebexTeamsMode.BOT
            self.token = None

            # Bot access token: prefer 'access_token', fall back to 'token'
            _at = access_token or token
            if not _at:
                msg = "A Webex Teams bot access token must be specified."
                self.logger.warning(msg)
                raise TypeError(msg)

            self.access_token = _at

            self.targets = parse_list(targets)

    def send(
        self,
        body,
        title="",
        notify_type=NotifyType.INFO,
        attach=None,
        **kwargs,
    ):
        """Perform Webex Teams Notification."""
        if self.mode == WebexTeamsMode.WEBHOOK:
            if attach and self.attachment_support:
                self.logger.warning(
                    "Webex Teams Webhooks do not support"
                    " attachments; use bot mode."
                )
            return self._send_webhook(body)

        return self._send_bot(body, attach=attach)

    def _send_webhook(self, body):
        """Post via incoming webhook (no attachment support)."""

        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json",
        }

        url = f"{self.notify_url}/{self.token}"

        payload = {
            (
                "markdown"
                if (self.notify_format == NotifyFormat.MARKDOWN)
                else "text"
            ): body,
        }

        self.logger.debug(
            "Webex Teams Webhook POST URL:"
            f" {url} (cert_verify={self.verify_certificate!r})"
        )
        self.logger.debug(f"Webex Teams Payload: {payload!s}")

        # Always call throttle before any remote server i/o is made
        self.throttle()
        try:
            r = requests.post(
                url,
                data=dumps(payload),
                headers=headers,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )
            if r.status_code not in (
                requests.codes.ok,
                requests.codes.no_content,
            ):
                status_str = NotifyWebexTeams.http_response_code_lookup(
                    r.status_code, WEBEX_HTTP_ERROR_MAP
                )

                self.logger.warning(
                    "Failed to send Webex Teams notification: "
                    "{}{}error={}.".format(
                        status_str,
                        ", " if status_str else "",
                        r.status_code,
                    )
                )

                self.logger.debug(
                    "Response Details:\r\n%r",
                    (r.content or b"")[:2000],
                )

                return False

            self.logger.info("Sent Webex Teams notification.")

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred sending Webex Teams notification."
            )
            self.logger.debug(f"Socket Exception: {e!s}")
            return False

        return True

    def _send_bot(self, body, attach=None):
        """Post via Bot/API to one or more rooms (supports attachments)."""

        if not self.targets:
            self.logger.warning(
                "Webex Teams Bot mode has no room IDs to notify, aborting."
            )
            return False

        has_error = False
        for room_id in self.targets:
            if not self._post_to_room(body, room_id, attach=attach):
                has_error = True

        return not has_error

    def _post_to_room(self, body, room_id, attach=None):
        """Send a single message (and optional attachments) to a room."""

        headers = {
            "User-Agent": self.app_id,
            "Authorization": f"Bearer {self.access_token}",
        }

        text_key = (
            "markdown"
            if self.notify_format == NotifyFormat.MARKDOWN
            else "text"
        )

        has_attachment = attach and self.attachment_support and len(attach) > 0

        if not has_attachment:
            # --- Text-only message sent as JSON ---
            headers["Content-Type"] = "application/json"
            payload = {
                "roomId": room_id,
                text_key: body,
            }

            self.logger.debug(
                "Webex Teams Bot POST URL:"
                f" {self.api_url}"
                f" (cert_verify={self.verify_certificate!r})"
            )
            self.logger.debug(f"Webex Teams Bot Payload: {payload!s}")

            self.throttle()
            try:
                r = requests.post(
                    self.api_url,
                    data=dumps(payload),
                    headers=headers,
                    verify=self.verify_certificate,
                    timeout=self.request_timeout,
                )
                if r.status_code != requests.codes.ok:
                    status_str = NotifyWebexTeams.http_response_code_lookup(
                        r.status_code, WEBEX_HTTP_ERROR_MAP
                    )
                    self.logger.warning(
                        "Failed to send Webex Teams Bot"
                        " notification:"
                        " {}{}error={}.".format(
                            status_str,
                            ", " if status_str else "",
                            r.status_code,
                        )
                    )
                    self.logger.debug(
                        "Response Details:\r\n%r",
                        (r.content or b"")[:2000],
                    )
                    return False

                self.logger.info(
                    f"Sent Webex Teams Bot notification to room {room_id}."
                )

            except requests.RequestException as e:
                self.logger.warning(
                    "A Connection error occurred sending"
                    " Webex Teams Bot notification."
                )
                self.logger.debug(f"Socket Exception: {e!s}")
                return False

            return True

        # --- Multipart attachment(s) ---
        for no, attachment in enumerate(attach, start=1):
            if not attachment:
                self.logger.error(
                    "Could not access Webex Teams attachment"
                    f" {attachment.url(privacy=True)}."
                )
                return False

            self.logger.debug(
                "Posting Webex Teams attachment"
                f" {attachment.url(privacy=True)}"
            )

            try:
                # open() returns a BytesIO for memory attachments; the
                # context manager guarantees the handle is closed on exit
                with attachment.open() as fp:
                    files = {
                        "files": (
                            (
                                attachment.name
                                if attachment.name
                                else f"file{no:03}.dat"
                            ),
                            fp,
                            attachment.mimetype,
                        ),
                    }

                    data = {"roomId": room_id}
                    # Include message body only with the first attachment
                    if no == 1:
                        data[text_key] = body

                    self.logger.debug(
                        "Webex Teams Bot attachment POST URL:"
                        f" {self.api_url}"
                        f" (cert_verify={self.verify_certificate!r})"
                    )

                    self.throttle()
                    r = requests.post(
                        self.api_url,
                        data=data,
                        headers=headers,
                        files=files,
                        verify=self.verify_certificate,
                        timeout=self.request_timeout,
                    )

                if r.status_code != requests.codes.ok:
                    status_str = NotifyWebexTeams.http_response_code_lookup(
                        r.status_code, WEBEX_HTTP_ERROR_MAP
                    )
                    self.logger.warning(
                        "Failed to send Webex Teams attachment"
                        " {}: {}{}error={}.".format(
                            attachment.name,
                            status_str,
                            ", " if status_str else "",
                            r.status_code,
                        )
                    )
                    self.logger.debug(
                        "Response Details:\r\n%r",
                        (r.content or b"")[:2000],
                    )
                    return False

                self.logger.info(
                    "Sent Webex Teams attachment"
                    f" {attachment.name} to room {room_id}."
                )

            except requests.RequestException as e:
                self.logger.warning(
                    "A Connection error occurred posting"
                    " Webex Teams attachment."
                )
                self.logger.debug(f"Socket Exception: {e!s}")
                return False

            except OSError as e:
                self.logger.warning(
                    "An I/O error occurred while reading {}.".format(
                        attachment.name if attachment.name else "attachment"
                    )
                )
                self.logger.debug(f"I/O Exception: {e!s}")
                return False

        return True

    @property
    def body_maxlen(self):
        """The maximum allowable characters allowed in the body per message.
        Webhook mode is limited to 1000 chars; the Bot API allows 7439."""
        return 1000 if self.mode == WebexTeamsMode.WEBHOOK else 7439

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        if self.mode == WebexTeamsMode.WEBHOOK:
            return (self.secure_protocol[0], self.token)

        # BOT mode
        return (self.secure_protocol[0], self.access_token)

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        params = {"mode": self.mode}
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        if self.mode == WebexTeamsMode.WEBHOOK:
            return "{schema}://{token}/?{params}".format(
                schema=self.secure_protocol[0],
                token=self.pprint(self.token, privacy, safe=""),
                params=NotifyWebexTeams.urlencode(params),
            )

        # BOT mode
        return "{schema}://{token}/{targets}/?{params}".format(
            schema=self.secure_protocol[0],
            token=self.pprint(self.access_token, privacy, safe=""),
            targets="/".join(
                [NotifyWebexTeams.quote(r, safe="") for r in self.targets]
            ),
            params=NotifyWebexTeams.urlencode(params),
        )

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us
        to re-instantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            return results

        # Explicit mode parameter wins
        if "mode" in results["qsd"] and results["qsd"]["mode"]:
            results["mode"] = NotifyWebexTeams.unquote(results["qsd"]["mode"])

        # Pull additional room-ID path entries (bot mode)
        entries = NotifyWebexTeams.split_path(results["fullpath"])

        # Support ?to= for room IDs
        if "to" in results["qsd"] and results["qsd"]["to"]:
            entries += NotifyWebexTeams.split_path(
                NotifyWebexTeams.unquote(results["qsd"]["to"])
            )

        # Support ?token= for the primary token/access_token
        if "token" in results["qsd"] and results["qsd"]["token"]:
            host = NotifyWebexTeams.unquote(results["qsd"]["token"])
        else:
            host = NotifyWebexTeams.unquote(results["host"])

        # Determine whether this is webhook or bot mode.
        # Path entries mean bot mode (room IDs present).
        # Explicit mode= has already been stored above.
        explicit_mode = results.get("mode", "")
        if explicit_mode == WebexTeamsMode.BOT or (
            entries and explicit_mode != WebexTeamsMode.WEBHOOK
        ):
            results["access_token"] = host
            results["targets"] = entries
        else:
            results["token"] = host

        return results

    @staticmethod
    def parse_native_url(url):
        """
        Support:
          https://api.ciscospark.com/v1/webhooks/incoming/WEBHOOK_TOKEN
          https://webexapis.com/v1/webhooks/incoming/WEBHOOK_TOKEN
        """

        result = re.match(
            r"^https?://(api\.ciscospark\.com|webexapis\.com)"
            r"/v[1-9][0-9]*/webhooks/incoming/"
            r"(?P<webhook_token>[A-Z0-9_-]+)/?"
            r"(?P<params>\?.+)?$",
            url,
            re.I,
        )

        if result:
            return NotifyWebexTeams.parse_url(
                "{schema}://{webhook_token}/{params}".format(
                    schema=NotifyWebexTeams.secure_protocol[0],
                    webhook_token=result.group("webhook_token"),
                    params=(
                        ""
                        if not result.group("params")
                        else result.group("params")
                    ),
                )
            )

        return None



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Simple API Reference:
#  - https://sendpulse.com/integrations/api/smtp

import base64
from email.utils import formataddr
from json import dumps, loads
import logging
import re

import requests

from .. import exception
from ..common import NotifyFormat, NotifyType, PersistentStoreMode
from ..conversion import convert_between
from ..locale import gettext_lazy as _
from ..utils.parse import is_email, parse_emails, validate_regex
from ..utils.sanitize import sanitize_payload
from .base import NotifyBase


class NotifySendPulse(NotifyBase):
    """
    A wrapper for Notify SendPulse Notifications
    """

    # The default descriptive name associated with the Notification
    service_name = "SendPulse"

    # The services URL
    service_url = "https://sendpulse.com"

    # The default secure protocol
    secure_protocol = "sendpulse"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://appriseit.com/services/sendpulse/"

    # Default to markdown
    notify_format = NotifyFormat.HTML

    # The default Email API URL to use
    notify_email_url = "https://api.sendpulse.com/smtp/emails"

    # Our OAuth Query
    notify_oauth_url = "https://api.sendpulse.com/oauth/access_token"

    # Support attachments
    attachment_support = True

    # Allow 300 requests per minute.
    # 60/300 = 0.2
    request_rate_per_sec = 0.2

    # Our default is to no not use persistent storage beyond in-memory
    # reference
    storage_mode = PersistentStoreMode.AUTO

    # Token expiry if not detected in seconds (below is 1 hr)
    token_expiry = 3600

    # The number of seconds to grace for early token renewal
    # Below states that 10 seconds bfore our token expiry, we'll
    # attempt to renew it
    token_expiry_edge = 10

    # Support attachments
    attachment_support = True

    # The default subject to use if one isn't specified.
    default_empty_subject = "<no subject>"

    # Define object templates
    templates = (
        "{schema}://{user}@{host}/{client_secret}/",
        "{schema}://{user}@{host}/{client_id}/{client_secret}/{targets}",
    )

    # Define our template arguments
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "user": {
                "name": _("User Name"),
                "type": "string",
                "required": True,
            },
            "host": {
                "name": _("Domain"),
                "type": "string",
                "required": True,
            },
            "client_id": {
                "name": _("Client ID"),
                "type": "string",
                "required": True,
                "private": True,
                "regex": (r"^[A-Z0-9._-]+$", "i"),
            },
            "client_secret": {
                "name": _("Client Secret"),
                "type": "string",
                "required": True,
                "private": True,
                "regex": (r"^[A-Z0-9._-]+$", "i"),
            },
            "target_email": {
                "name": _("Target Email"),
                "type": "string",
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "from": {
                "name": _("From Email"),
                "type": "string",
                "map_to": "from_addr",
            },
            "template": {
                # The template ID is an integer
                "name": _("Template ID"),
                "type": "int",
            },
            "id": {
                "alias_of": "client_id",
            },
            "secret": {
                "alias_of": "client_secret",
            },
            "to": {
                "alias_of": "targets",
            },
            "cc": {
                "name": _("Carbon Copy"),
                "type": "list:string",
            },
            "bcc": {
                "name": _("Blind Carbon Copy"),
                "type": "list:string",
            },
        },
    )

    # Support Template Dynamic Variables (Substitutions)
    template_kwargs = {
        "template_data": {
            "name": _("Template Data"),
            "prefix": "+",
        },
    }

    def __init__(
        self,
        client_id,
        client_secret,
        from_addr=None,
        targets=None,
        cc=None,
        bcc=None,
        template=None,
        template_data=None,
        **kwargs,
    ):
        """
        Initialize Notify SendPulse Object
        """
        super().__init__(**kwargs)

        # For tracking our email -> name lookups
        self.names = {}

        # Temporary from_addr to work with for parsing
        from_addr_ = [self.app_id, ""]

        if self.user:
            if self.host:
                # Prepare the bases of our email
                from_addr_ = [
                    from_addr_[0],
                    "{}@{}".format(
                        re.split(r"[\s@]+", self.user)[0],
                        self.host,
                    ),
                ]

            else:
                result = is_email(self.user)
                if result:
                    # Prepare the bases of our email and include domain
                    self.host = result["domain"]
                    from_addr_ = [
                        result["name"] if result["name"] else from_addr_[0],
                        self.user,
                    ]

        if isinstance(from_addr, str):
            result = is_email(from_addr)
            if result:
                from_addr_ = (
                    result["name"] if result["name"] else from_addr_[0],
                    result["full_email"],
                )
            else:
                # Only update the string but use the already detected info
                from_addr_[0] = from_addr

        result = is_email(from_addr_[1])
        if not result:
            # Parse Source domain based on from_addr
            msg = "Invalid ~From~ email specified: {}".format(
                "{} <{}>".format(from_addr_[0], from_addr_[1])
                if from_addr_[0]
                else "{}".format(from_addr_[1])
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # Store our lookup
        self.from_addr = from_addr_[1]
        self.names[from_addr_[1]] = from_addr_[0]

        # Client ID
        self.client_id = validate_regex(
            client_id, *self.template_tokens["client_id"]["regex"]
        )
        if not self.client_id:
            msg = "An invalid SendPulse Client ID ({}) was specified.".format(
                client_id
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # Client Secret
        self.client_secret = validate_regex(
            client_secret, *self.template_tokens["client_secret"]["regex"]
        )
        if not self.client_secret:
            msg = (
                "An invalid SendPulse Client Secret "
                "({}) was specified.".format(client_secret)
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # Acquire Targets (To Emails)
        self.targets = []

        # Acquire Carbon Copies
        self.cc = set()

        # Acquire Blind Carbon Copies
        self.bcc = set()

        # No template
        self.template = None
        if template:
            try:
                # Store our template
                self.template = int(template)

            except (TypeError, ValueError):
                # Not a valid integer; ignore entry
                err = (
                    "The SendPulse Template ID specified"
                    f" ({template}) is invalid."
                )
                self.logger.warning(err)
                raise TypeError(err) from None

        # Now our dynamic template data (if defined)
        self.template_data = (
            template_data if isinstance(template_data, dict) else {}
        )

        if targets:
            # Validate recipients (to:) and drop bad ones:
            for recipient in parse_emails(targets):
                result = is_email(recipient)
                if result:
                    self.targets.append(result["full_email"])
                    if result["name"]:
                        self.names[result["full_email"]] = result["name"]
                    continue

                self.logger.warning(
                    "Dropped invalid To email ({}) specified.".format(
                        recipient
                    ),
                )

        else:
            # If our target email list is empty we want to add ourselves to it
            self.targets.append(self.from_addr)

        # Validate recipients (cc:) and drop bad ones:
        for recipient in parse_emails(cc):
            result = is_email(recipient)
            if result:
                self.cc.add(result["full_email"])
                if result["name"]:
                    self.names[result["full_email"]] = result["name"]
                continue

            self.logger.warning(
                "Dropped invalid Carbon Copy email ({}) specified.".format(
                    recipient
                ),
            )

        # Validate recipients (bcc:) and drop bad ones:
        for recipient in parse_emails(bcc):
            result = is_email(recipient)
            if result:
                self.bcc.add(result["full_email"])
                if result["name"]:
                    self.names[result["full_email"]] = result["name"]
                continue

            self.logger.warning(
                "Dropped invalid Blind Carbon Copy email "
                "({}) specified.".format(recipient),
            )

        if len(self.targets) == 0:
            # Notify ourselves
            self.targets.append(self.from_addr)

        return

    @property
    def url_identifier(self):
        """
        Returns all of the identifiers that make this URL unique from
        another simliar one. Targets or end points should never be identified
        here.
        """
        return (self.secure_protocol, self.client_id, self.client_secret)

    def url(self, privacy=False, *args, **kwargs):
        """
        Returns the URL built dynamically based on specified arguments.
        """

        # Our URL parameters
        params = self.url_parameters(privacy=privacy, *args, **kwargs)

        if len(self.cc) > 0:
            # Handle our Carbon Copy Addresses
            params["cc"] = ",".join(
                [
                    formataddr(
                        (self.names.get(e, False), e),
                        # Swap comma for its escaped url code (if
                        # detected) since we use it as a delimiter
                        charset="utf-8",
                    ).replace(",", "%2C")
                    for e in self.cc
                ]
            )

        if len(self.bcc) > 0:
            # Handle our Blind Carbon Copy Addresses
            params["bcc"] = ",".join(
                [
                    formataddr(
                        (self.names.get(e, False), e),
                        # Swap comma for its escaped url code (if
                        # detected) since we use it as a delimiter
                        charset="utf-8",
                    ).replace(",", "%2C")
                    for e in self.bcc
                ]
            )

        if self.template:
            # Handle our Template ID if if was specified
            params["template"] = self.template

        # handle from=
        if self.names[self.from_addr] != self.app_id:
            params["from"] = self.names[self.from_addr]

        # Append our template_data into our parameter list
        params.update(
            {"+{}".format(k): v for k, v in self.template_data.items()}
        )

        # a simple boolean check as to whether we display our target emails
        # or not
        has_targets = not (
            len(self.targets) == 1 and self.targets[0] == self.from_addr
        )

        return "{schema}://{source}/{cid}/{secret}/{targets}?{params}".format(
            schema=self.secure_protocol,
            source=self.from_addr,
            cid=self.pprint(self.client_id, privacy, safe=""),
            secret=self.pprint(self.client_secret, privacy, safe=""),
            targets=""
            if not has_targets
            else "/".join(
                [NotifySendPulse.quote(x, safe="") for x in self.targets]
            ),
            params=NotifySendPulse.urlencode(params),
        )

    def __len__(self):
        """
        Returns the number of targets associated with this notification
        """
        return len(self.targets)

    def login(self):
        """
        Authenticates with the server to get a access_token
        """
        self.store.clear("access_token")
        payload = {
            "grant_type": "client_credentials",
            "client_id": self.client_id,
            "client_secret": self.client_secret,
        }

        success, response = self._fetch(self.notify_oauth_url, payload)
        if not success:
            return False

        access_token = response.get("access_token")

        # If we get here, we're authenticated
        try:
            expires = int(response.get("expires_in")) - self.token_expiry_edge
            if expires <= self.token_expiry_edge:
                self.logger.error(
                    "SendPulse token expiry limit returned was invalid"
                )
                return False

            elif expires > self.token_expiry:
                self.logger.warning(
                    "SendPulse token expiry limit fixed to: {}s".format(
                        self.token_expiry
                    )
                )
                expires = self.token_expiry - self.token_expiry_edge

        except (AttributeError, TypeError, ValueError):
            # expires_in was not an integer
            self.logger.warning(
                "SendPulse token expiry limit presumed to be: {}s".format(
                    self.token_expiry
                )
            )
            expires = self.token_expiry - self.token_expiry_edge

        self.store.set("access_token", access_token, expires=expires)

        return access_token

    def send(
        self,
        body,
        title="",
        notify_type=NotifyType.INFO,
        attach=None,
        **kwargs,
    ):
        """
        Perform SendPulse Notification
        """

        access_token = self.store.get("access_token") or self.login()
        if not access_token:
            return False

        # error tracking (used for function return)
        has_error = False

        # A Simple Email Payload Template
        payload_ = {
            "email": {
                "from": {
                    "name": self.names[self.from_addr],
                    "email": self.from_addr,
                },
                # To is populated further on
                "to": [],
                # A subject is a requirement, so if none is specified we must
                # set a default with at least 1 character or SendPulse will
                # deny our request
                "subject": title if title else self.default_empty_subject,
            }
        }

        # Prepare Email Message
        if self.notify_format == NotifyFormat.HTML:
            # HTML
            payload_["email"].update(
                {
                    "text": convert_between(
                        NotifyFormat.HTML, NotifyFormat.TEXT, body
                    ),
                    "html": base64.b64encode(body.encode("utf-8")).decode(
                        "ascii"
                    ),
                }
            )

        else:  # Text
            payload_["email"]["text"] = body

        if attach and self.attachment_support:
            attachments = {}

            # Send our attachments
            for no, attachment in enumerate(attach, start=1):
                # Perform some simple error checking
                if not attachment:
                    # We could not access the attachment
                    self.logger.error(
                        "Could not access SendPulse attachment {}.".format(
                            attachment.url(privacy=True)
                        )
                    )
                    return False

                try:
                    attachments[
                        attachment.name
                        if attachment.name
                        else f"file{no:03}.dat"
                    ] = attachment.base64()

                except exception.AppriseException:
                    # We could not access the attachment
                    self.logger.error(
                        "Could not access SendPulse attachment {}.".format(
                            attachment.url(privacy=True)
                        )
                    )
                    return False

                self.logger.debug(
                    "Appending SendPulse attachment {}".format(
                        attachment.url(privacy=True)
                    )
                )

            # Append our attachments to the payload
            payload_["email"].update(
                {
                    "attachments_binary": attachments,
                }
            )

        if self.template:
            payload_["email"].update(
                {
                    "template": {
                        "id": self.template,
                        "variables": self.template_data,
                    }
                }
            )

        targets = list(self.targets)
        while len(targets) > 0:
            target = targets.pop(0)

            # Create a copy of our template
            payload = payload_.copy()

            # the cc, bcc, to field must be unique or SendMail will fail, the
            # below code prepares this by ensuring the target isn't in the cc
            # list or bcc list. It also makes sure the cc list does not contain
            # any of the bcc entries
            cc = self.cc - self.bcc - {target}
            bcc = self.bcc - {target}

            #
            # prepare our 'to'
            #
            to = {"email": target}
            if target in self.names:
                to["name"] = self.names[target]

            # Set our target
            payload["email"]["to"] = [to]

            if len(cc):
                payload["email"]["cc"] = []
                for email in cc:
                    item = {
                        "email": email,
                    }
                if email in self.names:
                    item["name"] = self.names[email]

                payload["email"]["cc"].append(item)

            if len(bcc):
                payload["email"]["bcc"] = []
                for email in bcc:
                    item = {
                        "email": email,
                    }
                if email in self.names:
                    item["name"] = self.names[email]

                payload["email"]["bcc"].append(item)

            # Perform our post
            success, _response = self._fetch(
                self.notify_email_url, payload, target, retry=1
            )
            if not success:
                has_error = True
                continue

        return not has_error

    def _fetch(self, url, payload, target=None, retry=0):
        """
        Wrapper to request.post() to manage it's response better and make
        the send() function cleaner and easier to maintain.

        This function returns True if the _post was successful and False
        if it wasn't.
        """
        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json",
        }

        access_token = self.store.get("access_token")
        if access_token:
            headers.update({"Authorization": f"Bearer {access_token}"})

        # Some Debug Logging
        if self.logger.isEnabledFor(logging.DEBUG):
            # Due to attachments; output can be quite heavy and io intensive
            # To accommodate this, we only show our debug payload information
            # if required.
            self.logger.debug(
                f"SendPulse POST URL: {url}"
                f"(cert_verify={self.verify_certificate!r})"
            )
            self.logger.debug(
                "SendPulse Payload: %s", sanitize_payload(payload)
            )

        # Prepare our default response
        response = {}

        # Always call throttle before any remote server i/o is made
        self.throttle()
        try:
            r = requests.post(
                url,
                data=dumps(payload),
                headers=headers,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )

            try:
                response = loads(r.content)

            except (AttributeError, TypeError, ValueError):
                # This gets thrown if we can't parse our JSON Response
                #  - ValueError = r.content is Unparsable
                #  - TypeError = r.content is None
                #  - AttributeError = r is None
                self.logger.warning("Invalid response from SendPulse server.")
                self.logger.debug(
                    "Response Details:\r\n%r", (r.content or b"")[:2000]
                )
                return (False, {})

            # Reference status code
            status_code = r.status_code

            # Key likely expired, we'll reset it and try one more time
            if (
                status_code == requests.codes.unauthorized
                and retry
                and self.login()
            ):
                return self._fetch(url, payload, target, retry=retry - 1)

            if status_code not in (requests.codes.ok, requests.codes.accepted):
                # We had a problem
                status_str = NotifySendPulse.http_response_code_lookup(
                    status_code
                )

                if target:
                    self.logger.warning(
                        "Failed to send SendPulse notification to {}: "
                        "{}{}error={}.".format(
                            target,
                            status_str,
                            ", " if status_str else "",
                            status_code,
                        )
                    )
                else:
                    self.logger.warning(
                        "SendPulse Authentication Request failed: "
                        "{}{}error={}.".format(
                            status_str, ", " if status_str else "", status_code
                        )
                    )

                self.logger.debug(
                    "Response Details:\r\n%r", (r.content or b"")[:2000]
                )

            else:
                if target:
                    self.logger.info(
                        "Sent SendPulse notification to {}.".format(target)
                    )
                else:
                    self.logger.debug("SendPulse authentication successful")

                return (True, response)

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred sending SendPulse "
                "notification to {}.".format(target)
            )
            self.logger.debug("Socket Exception: {}".format(str(e)))

        return (False, response)

    @staticmethod
    def parse_url(url):
        """
        Parses the URL and returns enough arguments that can allow
        us to re-instantiate this object.

        """

        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Define our minimum requirements; defining them now saves us from
        # having to if/else all kinds of branches below...
        results["from_addr"] = None
        results["client_id"] = None
        results["client_secret"] = None

        # Prepare our targets
        results["targets"] = []

        # Our URL looks like this:
        #    {schema}://{from_addr}:{client_id}/{client_secret}/{targets}
        #
        # which actually equates to:
        #    {schema}://{user}@{host}/{client_id}/{client_secret}
        #                                  /{email1}/{email2}/etc..
        #                 ^       ^
        #                 |       |
        #                -from addr-
        if "from" in results["qsd"]:
            results["from_addr"] = NotifySendPulse.unquote(
                results["qsd"]["from"].rstrip()
            )

            if is_email(results["from_addr"]):
                # Our hostname is free'd up to be interpreted as part of the
                # targets
                results["targets"].append(
                    NotifySendPulse.unquote(results["host"])
                )
                results["host"] = ""

        if "user" in results["qsd"] and is_email(
            NotifySendPulse.unquote(results["user"])
        ):
            # Our hostname is free'd up to be interpreted as part of the
            # targets
            results["targets"].append(NotifySendPulse.unquote(results["host"]))
            results["host"] = ""

        # Get our potential email targets
        # First 2 elements are the client_id and client_secret
        results["targets"] += NotifySendPulse.split_path(results["fullpath"])
        # check for our client id
        if "id" in results["qsd"] and len(results["qsd"]["id"]):
            # Store our Client ID
            results["client_id"] = NotifySendPulse.unquote(
                results["qsd"]["id"]
            )

        elif results["targets"]:
            # Store our Client ID
            results["client_id"] = results["targets"].pop(0)

        if "secret" in results["qsd"] and len(results["qsd"]["secret"]):
            # Store our Client Secret
            results["client_secret"] = NotifySendPulse.unquote(
                results["qsd"]["secret"]
            )

        elif results["targets"]:
            # Store our Client Secret
            results["client_secret"] = results["targets"].pop(0)

        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"].append(
                NotifySendPulse.unquote(results["qsd"]["to"])
            )

        # Handle Carbon Copy Addresses
        if "cc" in results["qsd"] and len(results["qsd"]["cc"]):
            results["cc"] = NotifySendPulse.unquote(results["qsd"]["cc"])

        # Handle Blind Carbon Copy Addresses
        if "bcc" in results["qsd"] and len(results["qsd"]["bcc"]):
            results["bcc"] = NotifySendPulse.unquote(results["qsd"]["bcc"])

        # Handle Blind Carbon Copy Addresses
        if "template" in results["qsd"] and len(results["qsd"]["template"]):
            results["template"] = NotifySendPulse.unquote(
                results["qsd"]["template"]
            )

        # Add any template substitutions
        results["template_data"] = results["qsd+"]

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""A minimal, self-contained Slixmpp adapter.

This module provides a wrapper to Slixmpp for Apprise.
"""

from __future__ import annotations

import contextlib
import logging
import re
import ssl
import threading
import time
from typing import Any, Callable, Optional

import certifi

from ...compat import dataclass_compat as dataclass
from .common import SECURE_MODES, SecureXMPPMode

# Default our global support flag
SLIXMPP_SUPPORT_AVAILABLE = False

try:
    import asyncio
    from concurrent.futures import TimeoutError as FuturesTimeoutError

    import slixmpp

    SLIXMPP_SUPPORT_AVAILABLE = True

except ImportError:
    # Slixmpp is not available if code reaches here
    slixmpp = None  # type: ignore[assignment]
    asyncio = None  # type: ignore[assignment]
    FuturesTimeoutError = Exception  # type: ignore[misc]


@dataclass(frozen=True, slots=True)
class XMPPConfig:
    """Connection configuration."""

    host: str
    port: int
    jid: str
    password: str
    secure: str = SecureXMPPMode.STARTTLS
    verify_certificate: bool = True


# ---------------------------------------------------------------------------
# Logging Bridge
# ---------------------------------------------------------------------------
LOGGING_ID = "apprise.xmpp"
_LOG_BRIDGE_LOCK = threading.Lock()
_LOG_BRIDGED = False


def bridge_slixmpp_logging() -> None:
    """Bridge Slixmpp logging into Apprise logging handlers.

    This is intentionally idempotent to prevent handler duplication when many
    notifications are sent within the same process.
    """
    global _LOG_BRIDGED

    if _LOG_BRIDGED:
        return

    with _LOG_BRIDGE_LOCK:
        if _LOG_BRIDGED:
            return

        apprise_logger = logging.getLogger("apprise")
        slix_logger = logging.getLogger("slixmpp")

        existing = {id(h) for h in slix_logger.handlers}
        for handler in apprise_logger.handlers:
            if id(handler) not in existing:
                slix_logger.addHandler(handler)
                existing.add(id(handler))

        slix_logger.setLevel(apprise_logger.getEffectiveLevel())

        # Prevent duplicates via propagation chains.
        slix_logger.propagate = False

        _LOG_BRIDGED = True


def _close_awaitable(obj: Any) -> None:
    """Best-effort close for coroutine-like objects.

    Some test patches raise before awaiting, leaving coroutines to be
    garbage collected and triggering runtime warnings.
    """
    close = getattr(obj, "close", None)
    if callable(close):
        with contextlib.suppress(Exception):
            close()


# ---------------------------------------------------------------------------
# Internal Slixmpp Client Factory
# ---------------------------------------------------------------------------


_CLIENT_SUBCLASS_CACHE: dict[int, type[Any]] = {}


def _get_client_subclass(base_cls: type[Any]) -> type[Any]:
    """Return (and cache) the internal client subclass for a given base class.

    The tests monkeypatch `xmpp_adapter.slixmpp.ClientXMPP`, so we must resolve
    the base class dynamically at runtime, not at import time. We still cache
    the derived subclass per base class identity to avoid repeated class
    creation overhead in production.
    """
    key = id(base_cls)
    cached = _CLIENT_SUBCLASS_CACHE.get(key)
    if cached is not None:
        return cached

    class _Client(base_cls):  # type: ignore[misc]
        """Internal Slixmpp client for both one-shot and keepalive flows."""

        def __init__(
            self,
            jid: str,
            password: str,
            *,
            oneshot: bool,
            logger: Optional[logging.Logger] = None,
            targets: Optional[list[(str, str)]] = None,
            subject: str = "",
            body: str = "",
            before_message: Optional[Callable[[], None]] = None,
            # Multi-User Chat
            want_muc: bool = False,
            nick: str = "",
            want_roster: bool = False,
            roster_timeout: float = 0.0,
            session_started_evt: Optional[asyncio.Event] = None,
            # type: ignore[name-defined]
        ) -> None:
            super().__init__(jid, password)

            # Store the logger so _log_task can find it
            global LOGGING_ID
            self.logger = logger or logging.getLogger(LOGGING_ID)

            # Behaviour
            self._oneshot = bool(oneshot)

            # Send payload (only used in oneshot mode)
            self._targets = list(targets or [])
            self._subject = subject
            self._body = body
            self._before_message = before_message

            # Roster behaviour (both modes)
            self._want_roster = bool(want_roster)
            self._roster_timeout = float(roster_timeout)

            # Multi-User Chat
            self._want_muc = bool(want_muc)
            self._nick = nick
            if self._want_muc:
                with contextlib.suppress(Exception):
                    self.register_plugin("xep_0045")

            # Keepalive coordination (keepalive mode only)
            self._session_started_evt = session_started_evt

            # State
            self._auth_failed = False

            self.add_event_handler("session_start", self._on_session_start)
            self.add_event_handler("failed_auth", self._failed_auth)
            self.add_event_handler("disconnected", self._disconnected)

            # Keep behaviour predictable and close quickly.
            self.auto_reconnect = False

        async def _session_start(
            self, *args: object, **kwargs: object
        ) -> None:
            try:
                with contextlib.suppress(Exception):
                    self.send_presence()

                if self._want_roster and self._roster_timeout > 0:
                    roster_coro = self.get_roster()
                    try:
                        await asyncio.wait_for(
                            roster_coro,
                            timeout=self._roster_timeout,
                        )
                    except Exception:
                        _close_awaitable(roster_coro)

                # One-shot mode sends messages immediately on session_start and
                # then disconnects. Keepalive mode just signals readiness.
                if self._oneshot:
                    for mtype, target in self._targets:
                        if self._before_message:
                            self._before_message()

                        if mtype == "groupchat":
                            nick = (
                                self._nick
                                or getattr(self.boundjid, "user", "")
                                or "apprise"
                            )
                            muc_coro = self.plugin["xep_0045"].join_muc(
                                target, nick
                            )
                            try:
                                await asyncio.wait_for(muc_coro, timeout=5.0)
                            except Exception:
                                _close_awaitable(muc_coro)

                        self.send_message(
                            mto=target,
                            msubject=self._subject,
                            mbody=self._body,
                            mtype=mtype,
                        )

            finally:
                if self._oneshot:
                    self.disconnect()

                elif self._session_started_evt is not None:
                    self._session_started_evt.set()

        def _failed_auth(self, *args: object, **kwargs: object) -> None:
            # Authentication failure is always a hard failure.
            self._auth_failed = True
            if self._session_started_evt is not None:
                self._session_started_evt.clear()
            self.disconnect()

        def _on_session_start(self, *args: object, **kwargs: object) -> Any:
            """Slixmpp event handler entrypoint.

            Must be synchronous. Also, we must never fall back to
            asyncio.create_task() because that can bind to the wrong
            loop, or no loop, and leak coroutines.
            """
            coro = self._session_start(*args, **kwargs)

            # Schedule on the event loop for both one-shot and keepalive.
            loop = getattr(self, "loop", None)

            # If the loop is missing or already closing, we MUST close the
            # coroutine immediately to prevent "never awaited" warnings.
            if loop is None or not loop.is_running():
                with contextlib.suppress(Exception):
                    coro.close()
                return None

            try:
                task = loop.create_task(coro)

                def _log_task(t: asyncio.Task[Any]) -> None:
                    if t.cancelled():
                        return

                    exc = t.exception()
                    if exc is not None:
                        self.logger.error("XMPP task failed: %s", exc)

                task.add_done_callback(_log_task)

            except Exception:
                # Fallback closure if loop.create_task fails
                with contextlib.suppress(Exception):
                    coro.close()

            return None

        def _disconnected(self, *args: object, **kwargs: object) -> None:
            if self._session_started_evt is not None:
                self._session_started_evt.clear()

    _CLIENT_SUBCLASS_CACHE[key] = _Client
    return _Client


def _build_client(*args: Any, **kwargs: Any) -> Any:
    return _get_client_subclass(slixmpp.ClientXMPP)(*args, **kwargs)


# ---------------------------------------------------------------------------
# Adapter
# ---------------------------------------------------------------------------


class SlixmppAdapter:
    """Send a message to one or more targets.

    When keepalive is False, process() performs a one-shot connect, send,
    disconnect.

    When keepalive is True, send_message() keeps a session alive across calls.
    The connection is closed only when close() is called or the instance is
    garbage collected.
    """

    # Define a Slixmpp reference version to prevent this tool from working
    # under non-supported versions
    _supported_version = (1, 10, 0)

    # Flag to control if we are enabled or not
    # effectively.. .is the dependent slixmpp library available to us
    # or not
    _enabled = SLIXMPP_SUPPORT_AVAILABLE

    def __init__(
        self,
        config: XMPPConfig,
        targets: list[(str, str)],
        subject: str,
        body: str,
        timeout: float = 30.0,
        roster: bool = False,
        before_message: Optional[Callable[[], None]] = None,
        keepalive: bool = False,
        want_muc: bool = False,
        default_nickname: Optional[str] = None,
    ) -> None:
        self.config, self.targets, self.subject, self.body = (
            config,
            targets,
            subject,
            body,
        )

        self.timeout = max(5.0, float(timeout))
        self.roster, self.before_message, self.keepalive = (
            roster,
            before_message,
            keepalive,
        )
        self._want_muc = want_muc

        global LOGGING_ID
        self.logger = logging.getLogger(LOGGING_ID)

        self.nickname = default_nickname or "apprise"

        bridge_slixmpp_logging()

        # Keepalive internals (only used when keepalive=True)
        self._state_lock = threading.RLock()
        self._closing = False
        self._thread: Optional[threading.Thread] = None
        self._loop: Optional[asyncio.AbstractEventLoop] = None
        # type: ignore[name-defined]
        self._client: Optional[slixmpp.ClientXMPP] = None
        # type: ignore[name-defined]
        self._loop_ready = threading.Event()

        # asyncio primitives created inside the loop thread
        self._connect_lock: Optional[asyncio.Lock] = None
        # type: ignore[name-defined]
        self._session_started: Optional[asyncio.Event] = None
        # type: ignore[name-defined]

    def __del__(self) -> None:
        """Best effort close for keepalive sessions."""
        with contextlib.suppress(Exception):
            self.close()

    @staticmethod
    def _ssl_context(verify: bool) -> ssl.SSLContext:
        ctx = ssl.create_default_context(cafile=certifi.where())
        if not verify:
            ctx.check_hostname = False
            ctx.verify_mode = ssl.CERT_NONE
        return ctx

    @staticmethod
    def _loop_tick(loop: asyncio.AbstractEventLoop) -> None:
        """Run one final loop tick, closing the coroutine on error."""
        tick = asyncio.sleep(0)
        try:
            loop.run_until_complete(tick)

        except Exception:
            _close_awaitable(tick)

    @staticmethod
    def _finalize_loop(loop: asyncio.AbstractEventLoop) -> None:
        """Best-effort loop shutdown to avoid resource warnings."""

        with contextlib.suppress(Exception):
            # Cancel any pending tasks
            tasks = asyncio.all_tasks(loop)
            for task in tasks:
                task.cancel()

            if tasks:
                loop.run_until_complete(
                    asyncio.gather(*tasks, return_exceptions=True)
                )

        # Give the loop one final tick to process cancellations
        SlixmppAdapter._loop_tick(loop)

        with contextlib.suppress(Exception):
            loop.stop()

        # Only attempt to shutdown generators if the loop is still open.
        # Otherwise, creating the coroutine without running it triggers a
        # RuntimeWarning.
        if not loop.is_closed():
            with contextlib.suppress(Exception):
                ag_coro = loop.shutdown_asyncgens()
                try:
                    loop.run_until_complete(ag_coro)
                except Exception:
                    _close_awaitable(ag_coro)

        # Detach loop from thread policy
        with contextlib.suppress(Exception):
            asyncio.set_event_loop(None)

        if not loop.is_closed():
            with contextlib.suppress(Exception):
                loop.close()

    def close(self) -> None:
        """Close any persistent connection and stop the keepalive worker."""
        with self._state_lock:
            self._closing = True

        loop, client, thread = self._loop, self._client, self._thread
        if loop is None or thread is None:
            return

        def _shutdown() -> None:
            try:
                if client is not None:
                    client.disconnect()
            finally:
                with contextlib.suppress(Exception):
                    loop.stop()

        with contextlib.suppress(Exception):
            loop.call_soon_threadsafe(_shutdown)

        # Give a moment to exit gracefully.
        thread.join(max(1.0, min(5.0, self.timeout / 2.0)))

        # If the worker is still alive, do not clear state.
        alive = getattr(thread, "is_alive", None)
        if callable(alive) and alive():
            return

        with self._state_lock:
            # Detach from any thread-local loop to avoid creating a new
            # loop implicitly (Python 3.12+ may warn about it).
            with contextlib.suppress(Exception):
                asyncio.set_event_loop(None)

            self._thread = None
            self._loop = None
            self._client = None
            self._connect_lock = None
            self._session_started = None

    # -----------------------------------------------------------------------
    # One-shot behaviour (no keepalive)
    # -----------------------------------------------------------------------

    def process(self) -> bool:
        """Send the message, always returning within timeout."""
        done = threading.Event()
        result: list[Optional[bool]] = [None]

        if not self._enabled:
            # We are not turned on
            return False

        shared: dict[str, Any] = {"loop": None, "client": None}

        def runner() -> None:
            loop: Optional[asyncio.AbstractEventLoop] = None
            # type: ignore[name-defined]
            start = time.monotonic()

            try:
                loop = asyncio.new_event_loop()  # type: ignore[union-attr]
                asyncio.set_event_loop(loop)  # type: ignore[union-attr]
                shared["loop"] = loop

                targets = (
                    list(self.targets)
                    if self.targets
                    else [("chat", self.config.jid)]
                )

                roster_timeout = (
                    max(2.0, min(10.0, self.timeout / 3.0))
                    if self.roster
                    else 0.0
                )

                client = _build_client(
                    jid=self.config.jid,
                    password=self.config.password,
                    oneshot=True,
                    logger=self.logger,
                    targets=targets,
                    subject=self.subject,
                    body=self.body,
                    before_message=self.before_message,
                    want_muc=self._want_muc,
                    nick=self.nickname,
                    want_roster=self.roster,
                    roster_timeout=roster_timeout,
                    session_started_evt=None,
                )

                shared["client"] = client

                # Prevent Slixmpp from owning loop lifecycle
                with contextlib.suppress(Exception):
                    client.loop = loop  # type: ignore[assignment]

                # Resolve connection behaviour from secure mode
                mode_cfg = SECURE_MODES.get(self.config.secure)
                if not mode_cfg:
                    raise ValueError(
                        f"Unsupported XMPP secure mode: {self.config.secure}"
                    )

                client.enable_plaintext = bool(mode_cfg["enable_plaintext"])
                client.enable_starttls = bool(mode_cfg["enable_starttls"])
                client.enable_direct_tls = bool(mode_cfg["enable_direct_tls"])

                # Only attach an SSL context when TLS may be used
                if not client.enable_plaintext:
                    client.ssl_context = self._ssl_context(
                        self.config.verify_certificate
                    )

                # Slixmpp >= 1.10.0 connect() returns a Future.
                connect_timeout = max(3.0, min(15.0, self.timeout / 3.0))
                connect_fut = client.connect(
                    host=self.config.host,
                    port=self.config.port,
                )

                try:
                    ok = loop.run_until_complete(
                        asyncio.wait_for(
                            connect_fut,
                            timeout=connect_timeout,
                        )
                    )
                    if ok is False:
                        self.logger.warning("XMPP connect failed.")
                        with contextlib.suppress(Exception):
                            client.disconnect()
                        result[0] = False
                        return

                except asyncio.TimeoutError:
                    self.logger.warning(
                        "XMPP connect timed out after %.2fs", connect_timeout
                    )
                    result[0] = False
                    return

                except Exception as e:
                    self.logger.debug("XMPP connect failed: %s", e)
                    result[0] = False
                    return

                # Run until disconnected, but still respect our overall
                # timeout.
                elapsed = time.monotonic() - start
                remaining = max(0.0, self.timeout - elapsed)
                run_timeout = max(1.0, remaining)

                try:
                    loop.run_until_complete(
                        asyncio.wait_for(  # type: ignore[arg-type]
                            client.disconnected, timeout=run_timeout
                        )
                    )
                except asyncio.TimeoutError:  # type: ignore[attr-defined]
                    self.logger.warning(
                        "XMPP session timed out after %.2fs", run_timeout
                    )
                    with contextlib.suppress(Exception):
                        client.disconnect()
                    result[0] = False
                    return

                # Disconnect happened, success depends on auth state
                result[0] = not bool(getattr(client, "_auth_failed", False))

            except Exception as e:  # pragma: no cover
                self.logger.warning("XMPP send failed.")
                self.logger.debug("XMPP Exception: %s", e)
                result[0] = False

            finally:
                loop = shared.get("loop")
                if loop is not None:
                    self._finalize_loop(loop)
                done.set()

        t = threading.Thread(target=runner, name="apprise-xmpp", daemon=True)
        t.start()

        if not done.wait(timeout=self.timeout):
            self.logger.warning(
                "XMPP send timed out after %.2fs.", self.timeout
            )
            result[0] = False

            loop_obj = shared.get("loop")
            client_obj = shared.get("client")

            if loop_obj is not None:
                loop = loop_obj  # type: ignore[assignment]
                try:
                    if client_obj is not None:
                        client = client_obj  # type: ignore[assignment]
                        loop.call_soon_threadsafe(client.disconnect)
                except Exception:
                    pass

                with contextlib.suppress(Exception):
                    loop.call_soon_threadsafe(loop.stop)

            t.join(timeout=0.25)

        return bool(result[0])

    # -----------------------------------------------------------------------
    # Keepalive Behaviour
    # -----------------------------------------------------------------------

    def _ensure_keepalive_worker(self) -> bool:
        """Ensure the background loop and client exist."""
        if not self.keepalive:
            return False

        with self._state_lock:
            if self._closing:
                return False

            if self._thread is not None and self._thread.is_alive():
                return True

            if not self._enabled:
                return False

            self._loop_ready.clear()

            self._thread = threading.Thread(
                target=self._keepalive_runner,
                name="apprise-xmpp-keepalive",
                daemon=True,
            )
            self._thread.start()

        if not self._loop_ready.wait(timeout=self.timeout):
            self.logger.warning(
                "XMPP keepalive worker failed to start within %.2fs",
                self.timeout,
            )
            return False

        return True

    def _keepalive_runner(self) -> None:
        loop: Optional[asyncio.AbstractEventLoop] = None
        # type: ignore[name-defined]
        published = False

        try:
            loop = asyncio.new_event_loop()  # type: ignore[union-attr]
            asyncio.set_event_loop(loop)  # type: ignore[union-attr]

            session_started = asyncio.Event()  # type: ignore[union-attr]
            connect_lock = asyncio.Lock()  # type: ignore[union-attr]

            roster_timeout = (
                max(2.0, min(10.0, self.timeout / 3.0)) if self.roster else 0.0
            )

            client = _build_client(
                jid=self.config.jid,
                password=self.config.password,
                oneshot=False,
                logger=self.logger,
                want_muc=self._want_muc,
                want_roster=self.roster,
                roster_timeout=roster_timeout,
                session_started_evt=session_started,
            )

            with contextlib.suppress(Exception):
                client.loop = loop  # type: ignore[assignment]

            mode_cfg = SECURE_MODES.get(self.config.secure)
            if not mode_cfg:
                raise ValueError(
                    f"Unsupported XMPP secure mode: {self.config.secure}"
                )

            client.enable_plaintext = bool(mode_cfg["enable_plaintext"])
            client.enable_starttls = bool(mode_cfg["enable_starttls"])
            client.enable_direct_tls = bool(mode_cfg["enable_direct_tls"])

            if not client.enable_plaintext:
                client.ssl_context = self._ssl_context(
                    self.config.verify_certificate
                )

            # keepalive=yes implies enabling XEP-0199 keepalive pings
            with contextlib.suppress(Exception):
                client.register_plugin("xep_0199", {"keepalive": True})

            if self._want_muc:
                # Multi-User Chat
                with contextlib.suppress(Exception):
                    client.register_plugin("xep_0045")

            with self._state_lock:
                if self._closing:
                    return

                self._loop = loop
                self._client = client
                self._connect_lock = connect_lock
                self._session_started = session_started
                published = True

            self._loop_ready.set()

            loop.run_forever()

        except Exception as e:  # pragma: no cover
            self.logger.warning("XMPP keepalive worker failed.")
            self.logger.debug("XMPP keepalive exception: %s", e)

        finally:
            if published:
                with self._state_lock:
                    if self._closing and self._loop is loop:
                        # Clear internal references if we are exiting the
                        # worker.
                        self._loop = None
                        self._client = None
                        self._connect_lock = None
                        self._session_started = None
                        self._thread = None

            if loop is not None:
                self._finalize_loop(loop)

    async def _connect_if_required(self) -> bool:
        if self._loop is None or self._client is None:
            return False
        if self._connect_lock is None or self._session_started is None:
            return False

        # If auth already failed, do not pretend a connection is ready.
        if bool(getattr(self._client, "_auth_failed", False)):
            return False

        async with self._connect_lock:
            if self._session_started.is_set():
                return True

            connect_timeout = max(3.0, min(15.0, self.timeout / 3.0))

            try:
                fut = self._client.connect(
                    host=self.config.host,
                    port=self.config.port,
                )
                connect_ok = await asyncio.wait_for(  # type: ignore[arg-type]
                    fut, timeout=connect_timeout
                )

                # honour boolean connect() result in keepalive.
                if not connect_ok:
                    self.logger.warning("XMPP connect failed.")
                    with contextlib.suppress(Exception):
                        self._client.disconnect()
                    return False

            except asyncio.TimeoutError:  # type: ignore[attr-defined]
                self.logger.warning(
                    "XMPP connect timed out after %.2fs", connect_timeout
                )
                return False

            except Exception as e:
                self.logger.debug("XMPP connect failed: %s", e)
                return False

            try:
                session_wait = self._session_started.wait()
                await asyncio.wait_for(
                    session_wait,
                    timeout=connect_timeout,
                )
            except asyncio.TimeoutError:  # type: ignore[attr-defined]
                _close_awaitable(session_wait)
                self.logger.warning(
                    "XMPP session did not start within %.2fs",
                    connect_timeout,
                )
                return False

            except Exception:
                _close_awaitable(session_wait)

                return False

            # If auth failed during startup, treat as failure.
            return not bool(getattr(self._client, "_auth_failed", False))

    async def _send_keepalive_async(
        self,
        targets: list[(str, str)],
        subject: str,
        body: str,
    ) -> bool:
        if self._client is None:
            return False

        ok = await self._connect_if_required()
        if ok is False:
            return False

        # Auth failed after connect, do not send.
        if bool(getattr(self._client, "_auth_failed", False)):
            return False

        send_targets = targets if targets else [("chat", self.config.jid)]

        try:
            for mtype, target in send_targets:
                if mtype == "groupchat":
                    nick = (
                        getattr(self._client.boundjid, "user", "")
                        or self.nickname
                    )
                    muc_coro = self._client.plugin["xep_0045"].join_muc(
                        target, nick
                    )
                    try:
                        await asyncio.wait_for(muc_coro, timeout=5.0)
                    except Exception:
                        _close_awaitable(muc_coro)

                self._client.send_message(
                    mto=target,
                    msubject=subject,
                    mbody=body,
                    mtype=mtype,
                )
            return True

        except Exception as e:
            self.logger.debug("XMPP send failed: %s", e)
            if self._session_started is not None:
                self._session_started.clear()
            return False

    def send_message(
        self,
        targets: Optional[list[(str, str)]] = None,
        subject: Optional[str] = None,
        body: Optional[str] = None,
    ) -> bool:
        """Send a message, keeping the session alive if keepalive=True."""

        if not self.keepalive:
            # Fallback to one-shot behaviour using current stored attributes
            if targets is not None:
                self.targets = targets
            if subject is not None:
                self.subject = subject
            if body is not None:
                self.body = body
            return self.process()

        if not self._ensure_keepalive_worker():
            return False

        loop = self._loop
        if loop is None:
            return False

        targets = self.targets if targets is None else targets
        subject = self.subject if subject is None else subject
        body = self.body if body is None else body

        coro = self._send_keepalive_async(
            targets=targets,
            subject=subject,
            body=body,
        )

        try:
            fut = asyncio.run_coroutine_threadsafe(  # type: ignore[union-attr]
                coro,
                loop,
            )
            return bool(fut.result(timeout=self.timeout))

        except FuturesTimeoutError:
            self.logger.warning(
                "XMPP keepalive send timed out after %.2fs", self.timeout
            )
            if self._session_started is not None:
                with contextlib.suppress(Exception):
                    loop.call_soon_threadsafe(self._session_started.clear)
            return False

        except Exception as e:
            with contextlib.suppress(Exception):
                coro.close()
            self.logger.debug("XMPP keepalive send exception: %s", e)
            return False

    @staticmethod
    def package_dependency() -> str:
        """Defines our static dependency for this adapter to work."""
        version = ".".join([str(v) for v in SlixmppAdapter._supported_version])
        return f"slixmpp >= {version}"

    @staticmethod
    def supported_version(version: Optional[str] = None) -> bool:
        """Returns true if we currently have a version of Slixmpp supported.

        Provided string describes a version in format of major.minor.patch.
        """
        if SLIXMPP_SUPPORT_AVAILABLE:
            m = re.match(
                r"^(?P<major>\d+)(\.(?P<minor>\d+)(\.(?P<patch>\d+))?)?",
                version or getattr(slixmpp, "__version__", "") or "",
            )
            if not m:
                return False

            return (
                int(m.group("major")),
                int(m.group("minor") or 0),
                int(m.group("patch") or 0),
            ) >= SlixmppAdapter._supported_version

        return False



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Steps to get your Postmark Server API Token:
#  1. Visit https://account.postmarkapp.com/ and sign in.
#  2. Select the server you wish to send from (or create a new one).
#  3. In the server settings, click "API Tokens".
#  4. Copy the Server API Token shown on that page.
#
#  Your sender address (From Email) must be a verified sender signature
#  or belong to a verified sending domain in Postmark.  Visit:
#    https://account.postmarkapp.com/signature_domains
#
#  Build your Apprise URL as follows:
#    postmark://{apikey}:{from_email}
#    postmark://{apikey}:{from_email}/{to_email}
#    postmark://{apikey}:{from_email}/{to_email1}/{to_email2}/{to_emailN}
#
#  Use the optional query parameters to add CC, BCC, Reply-To:
#    postmark://{apikey}:{from_email}?bcc=bcc@example.com
#    postmark://{apikey}:{from_email}?cc=cc@example.com
#    postmark://{apikey}:{from_email}?reply=reply@example.com
#    postmark://{apikey}:{from_email}?name=Display+Name
#
#  API Reference:
#    https://postmarkapp.com/developer/api/email-api

from email.utils import formataddr
from json import dumps
import logging

import requests

from .. import exception
from ..common import NotifyFormat, NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import is_email, parse_emails, validate_regex
from ..utils.sanitize import sanitize_payload
from .base import NotifyBase

# Provide some known codes Postmark uses and what they translate to.
# Reference: https://postmarkapp.com/developer/api/overview#error-codes
POSTMARK_HTTP_ERROR_MAP = {
    401: "Unauthorized - Invalid or missing Server API Token.",
    422: "Unprocessable Entity - Invalid payload or configuration.",
    429: "Too Many Requests - Rate limit exceeded.",
    500: "Internal Server Error.",
}


class NotifyPostmark(NotifyBase):
    """A wrapper for Postmark Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Postmark"

    # The services URL
    service_url = "https://postmarkapp.com/"

    # The default secure protocol
    secure_protocol = "postmark"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://appriseit.com/services/postmark/"

    # Default to HTML
    notify_format = NotifyFormat.HTML

    # The Postmark email API endpoint
    notify_url = "https://api.postmarkapp.com/email"

    # Support attachments
    attachment_support = True

    # Postmark practical send rate is ~50 messages/second.
    # 60 / 50 = 1.2
    request_rate_per_sec = 1.2

    # The default subject to use if one is not specified
    default_empty_subject = "<no subject>"

    # Define object URL templates
    templates = (
        "{schema}://{apikey}:{from_email}",
        "{schema}://{apikey}:{from_email}/{targets}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "apikey": {
                "name": _("API Key"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "from_email": {
                "name": _("Source Email"),
                "type": "string",
                "required": True,
            },
            "target_email": {
                "name": _("Target Email"),
                "type": "string",
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "name": {
                "name": _("From Name"),
                "type": "string",
                "map_to": "from_name",
            },
            "reply": {
                "name": _("Reply To Email"),
                "type": "list:string",
                "map_to": "reply_to",
            },
            "to": {
                "alias_of": "targets",
            },
            "cc": {
                "name": _("Carbon Copy"),
                "type": "list:string",
            },
            "bcc": {
                "name": _("Blind Carbon Copy"),
                "type": "list:string",
            },
        },
    )

    def __init__(
        self,
        apikey,
        from_email,
        targets=None,
        cc=None,
        bcc=None,
        from_name=None,
        reply_to=None,
        **kwargs,
    ):
        """Initialize Notify Postmark Object."""
        super().__init__(**kwargs)

        # API Key (Server Token)
        self.apikey = validate_regex(apikey)
        if not self.apikey:
            msg = f"An invalid Postmark API Key ({apikey}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        # Validate our from email address
        result = is_email(from_email)
        if not result:
            msg = (
                f"An invalid Postmark From email ({from_email}) was specified."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # Store our from address as a (name, email) tuple
        self.from_addr = (
            result["name"] if result["name"] is not None else False,
            result["full_email"],
        )

        # from_name overrides any name embedded in the from email string
        if from_name:
            self.from_addr = (from_name, self.from_addr[1])

        # For tracking email -> display name lookups
        self.names = {}

        # Acquire Targets (To Emails)
        self.targets = []

        # Acquire Carbon Copies
        self.cc = set()

        # Acquire Blind Carbon Copies
        self.bcc = set()

        # Acquire Reply-To addresses
        self.reply_to = set()

        # Validate recipients (to:) and drop bad ones:
        if targets:
            for recipient in parse_emails(targets):
                result = is_email(recipient)
                if result:
                    self.targets.append(result["full_email"])
                    # Index name if one exists
                    self.names[result["full_email"]] = (
                        result["name"] if result["name"] else False
                    )
                    continue

                self.logger.warning(
                    "Dropped invalid Postmark To email "
                    f"({recipient}) specified.",
                )

        else:
            # Default to the from address when no targets are specified
            self.targets.append(self.from_addr[1])

        # Validate recipients (cc:) and drop bad ones:
        for recipient in parse_emails(cc):
            result = is_email(recipient)
            if result:
                self.cc.add(result["full_email"])
                # Index name if one exists
                self.names[result["full_email"]] = (
                    result["name"] if result["name"] else False
                )
                continue

            self.logger.warning(
                "Dropped invalid Postmark Carbon Copy email "
                f"({recipient}) specified.",
            )

        # Validate recipients (bcc:) and drop bad ones:
        for recipient in parse_emails(bcc):
            result = is_email(recipient)
            if result:
                self.bcc.add(result["full_email"])
                continue

            self.logger.warning(
                "Dropped invalid Postmark Blind Carbon Copy "
                f"email ({recipient}) specified.",
            )

        # Validate reply-to addresses and drop bad ones:
        for recipient in parse_emails(reply_to):
            result = is_email(recipient)
            if result:
                self.reply_to.add(result["full_email"])
                # Index name if one exists
                self.names[result["full_email"]] = (
                    result["name"] if result["name"] else False
                )
                continue

            self.logger.warning(
                "Dropped invalid Postmark Reply To email "
                f"({recipient}) specified.",
            )

        return

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (self.secure_protocol, self.apikey, self.from_addr[1])

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Our URL parameters
        params = self.url_parameters(privacy=privacy, *args, **kwargs)

        # Include from display name if one is set
        if self.from_addr[0]:
            params["name"] = self.from_addr[0]

        if self.cc:
            # Handle our Carbon Copy Addresses
            params["cc"] = ",".join(
                [
                    formataddr(
                        (self.names.get(e, False), e),
                        charset="utf-8",
                    ).replace(",", "%2C")
                    for e in self.cc
                ]
            )

        if self.bcc:
            # Handle our Blind Carbon Copy Addresses
            params["bcc"] = ",".join(self.bcc)

        if self.reply_to:
            # Handle our Reply-To Addresses
            params["reply"] = ",".join(
                [
                    formataddr(
                        (self.names.get(e, False), e),
                        charset="utf-8",
                    ).replace(",", "%2C")
                    for e in self.reply_to
                ]
            )

        # Determine whether to display target emails in the URL
        has_targets = not (
            len(self.targets) == 1 and self.targets[0] == self.from_addr[1]
        )

        return "{schema}://{apikey}:{from_email}/{targets}?{params}".format(
            schema=self.secure_protocol,
            apikey=self.pprint(self.apikey, privacy, safe=""),
            # never encode email since it plays a role in our hostname
            from_email=self.from_addr[1],
            targets=(
                ""
                if not has_targets
                else "/".join(
                    [NotifyPostmark.quote(x, safe="@") for x in self.targets]
                )
            ),
            params=NotifyPostmark.urlencode(params),
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        return max(len(self.targets), 1)

    def send(
        self,
        body,
        title="",
        notify_type=NotifyType.INFO,
        attach=None,
        **kwargs,
    ):
        """Perform Postmark Notification."""

        if not self.targets:
            # There is no one to email; we're done
            self.logger.warning(
                "There are no Postmark email recipients to notify"
            )
            return False

        # Prepare our headers
        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json",
            "Accept": "application/json",
            "X-Postmark-Server-Token": self.apikey,
        }

        # error tracking (used for function return)
        has_error = False

        # Prepare the From field (with optional display name)
        from_field = formataddr(self.from_addr, charset="utf-8")

        # Base payload template (shared across all targets)
        payload_ = {
            "From": from_field,
            "Subject": (title if title else self.default_empty_subject),
        }

        # Set body content based on the notify format
        if self.notify_format == NotifyFormat.HTML:
            payload_["HtmlBody"] = body
        else:
            payload_["TextBody"] = body

        if attach and self.attachment_support:
            # Prepare attachment list
            attachments = []

            for no, attachment in enumerate(attach, start=1):
                # Perform some simple error checking
                if not attachment:
                    # We could not access the attachment
                    self.logger.error(
                        "Could not access Postmark attachment"
                        f" {attachment.url(privacy=True)}."
                    )
                    return False

                try:
                    # Append base64-encoded attachment to list
                    attachments.append(
                        {
                            "Name": (
                                attachment.name
                                if attachment.name
                                else f"file{no:03}.dat"
                            ),
                            "Content": attachment.base64(),
                            "ContentType": attachment.mimetype,
                        }
                    )

                except exception.AppriseException:
                    # We could not access the attachment
                    self.logger.error(
                        "Could not access Postmark attachment"
                        f" {attachment.url(privacy=True)}."
                    )
                    return False

                self.logger.debug(
                    "Appending Postmark attachment"
                    f" {attachment.url(privacy=True)}"
                )

            # Append our attachment list to the base payload
            payload_["Attachments"] = attachments

        # Iterate over each recipient target
        targets = list(self.targets)
        while len(targets) > 0:
            target = targets.pop(0)

            # Create a per-target copy of our base payload
            payload = payload_.copy()

            # Unique cc/bcc/reply-to management -- remove target from each
            cc = self.cc - self.bcc - {target}
            bcc = self.bcc - {target}
            reply_to = self.reply_to - {target}

            # Set our main recipient
            payload["To"] = target

            if cc:
                # Format CC addresses with optional display names
                payload["Cc"] = ",".join(
                    [
                        formataddr(
                            (self.names.get(a, False), a),
                            charset="utf-8",
                        )
                        for a in cc
                    ]
                )

            if bcc:
                # BCC addresses (plain, no display names)
                payload["Bcc"] = ",".join(bcc)

            if reply_to:
                # Format Reply-To addresses with optional display names
                payload["ReplyTo"] = ",".join(
                    [
                        formataddr(
                            (self.names.get(a, False), a),
                            charset="utf-8",
                        )
                        for a in reply_to
                    ]
                )

            # Some Debug Logging
            if self.logger.isEnabledFor(logging.DEBUG):
                # Due to attachments, output can be quite heavy; only
                # show the debug payload when debug logging is active.
                self.logger.debug(
                    "Postmark POST URL:"
                    f" {self.notify_url} "
                    f"(cert_verify={self.verify_certificate!r})"
                )
                self.logger.debug(
                    "Postmark Payload: %s", sanitize_payload(payload)
                )

            # Always call throttle before any remote server i/o is made
            self.throttle()

            try:
                r = requests.post(
                    self.notify_url,
                    data=dumps(payload),
                    headers=headers,
                    verify=self.verify_certificate,
                    timeout=self.request_timeout,
                )
                if r.status_code != requests.codes.ok:
                    # We had a problem
                    status_str = NotifyPostmark.http_response_code_lookup(
                        r.status_code, POSTMARK_HTTP_ERROR_MAP
                    )

                    self.logger.warning(
                        "Failed to send Postmark notification to "
                        "{}: {}{}error={}.".format(
                            target,
                            status_str,
                            ", " if status_str else "",
                            r.status_code,
                        )
                    )

                    self.logger.debug(
                        "Response Details:\r\n%r",
                        (r.content or b"")[:2000],
                    )

                    # Mark our failure
                    has_error = True
                    continue

                else:
                    self.logger.info(
                        f"Sent Postmark notification to {target}."
                    )

            except requests.RequestException as e:
                self.logger.warning(
                    "A Connection error occurred sending Postmark "
                    f"notification to {target}."
                )
                self.logger.debug(f"Socket Exception: {e!s}")

                # Mark our failure
                has_error = True
                continue

        return not has_error

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us
        to re-instantiate this object."""

        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Our URL looks like this:
        #    {schema}://{apikey}:{from_email}/{targets}
        #
        # which actually equates to:
        #    {schema}://{user}:{password}@{host}/{email1}/{email2}/etc
        #                  ^       ^         ^
        #                  |       |         |
        #               apikey    user     domain
        #                              (from email parts)

        # Handle apikey= query param override
        if "apikey" in results["qsd"] and results["qsd"]["apikey"]:
            results["apikey"] = NotifyPostmark.unquote(
                results["qsd"]["apikey"]
            )
        else:
            # Fall back to the user field
            results["apikey"] = NotifyPostmark.unquote(results["user"])

        # Our targets list
        results["targets"] = []

        # Handle from= query param (alternative from address)
        if "from" in results["qsd"] and results["qsd"]["from"]:
            results["from_email"] = NotifyPostmark.unquote(
                results["qsd"]["from"]
            )
            # Treat the host as a target when from= is given explicitly
            if results.get("host"):
                results["targets"].append(
                    NotifyPostmark.unquote(results["host"])
                )

        else:
            # Reconstruct from_email from {password}@{host}
            results["from_email"] = "{}@{}".format(
                NotifyPostmark.unquote(
                    results["password"]
                    if results["password"]
                    else results["user"]
                ),
                NotifyPostmark.unquote(results["host"]),
            )

        # Handle from display name
        if "name" in results["qsd"] and results["qsd"]["name"]:
            results["from_name"] = NotifyPostmark.unquote(
                results["qsd"]["name"]
            )

        # Acquire targets from the URL path
        results["targets"].extend(
            NotifyPostmark.split_path(results["fullpath"])
        )

        # Support ?to= for additional targets
        if "to" in results["qsd"] and results["qsd"]["to"]:
            results["targets"] += NotifyPostmark.parse_list(
                results["qsd"]["to"]
            )

        # Handle Carbon Copy Addresses
        if "cc" in results["qsd"] and results["qsd"]["cc"]:
            results["cc"] = NotifyPostmark.parse_list(results["qsd"]["cc"])

        # Handle Blind Carbon Copy Addresses
        if "bcc" in results["qsd"] and results["qsd"]["bcc"]:
            results["bcc"] = NotifyPostmark.parse_list(results["qsd"]["bcc"])

        # Handle Reply-To Addresses
        if "reply" in results["qsd"] and results["qsd"]["reply"]:
            results["reply_to"] = results["qsd"]["reply"]

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

#
# API: https://dot.mindreset.tech/docs/service/open/text_api
#      https://dot.mindreset.tech/docs/service/open/image_api
#
# New API endpoints (v2):
#   - Text API: /api/authV2/open/device/:deviceId/text
#   - Image API: /api/authV2/open/device/:deviceId/image
#
# Note: Old endpoints (/api/open/text and /api/open/image) are deprecated
# and will be removed in the future. Requests are automatically forwarded
# to the new endpoints.
#
# Text API Fields:
#   - refreshNow (bool, optional, default true): controls display timing.
#   - title (string, optional): title text shown on screen.
#   - message (string, optional): body text shown on screen.
#   - signature (string, optional): footer/signature text.
#   - icon (string, optional): base64 PNG icon (40px x 40px).
#   - link (string, optional): tap-to-interact target URL.
#   - taskKey (string, optional): specify which text API content to update.
#
# Image API Fields:
#   - refreshNow (bool, optional, default true): controls display timing.
#   - image (string, required): base64 PNG image (296px x 152px).
#   - link (string, optional): tap-to-interact target URL.
#   - border (number, optional, default 0): 0=white, 1=black frame.
#   - ditherType (string, optional, default DIFFUSION): dithering mode.
#   - ditherKernel (string, optional, default FLOYD_STEINBERG):
#     dithering kernel.
#   - taskKey (string, optional): specify which image API content to update.
#
# Mode selection:
#   - text (default): smart dual-send mode.  Body/title go to the Text API;
#     image= param or an attachment goes to the Image API.  When both are
#     present, text is dispatched first, then image.  If only one is
#     available, only that API is called.
#   - image: only the Image API is called; body and title are ignored.
#
#   Providing image= in the URL without an explicit mode= leaves the mode
#   as the default (text), which will send the image alongside any text.

from contextlib import suppress
import json
import logging

import requests

from ..common import NotifyImageSize, NotifyType
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from ..utils.parse import parse_bool
from ..utils.sanitize import sanitize_payload
from .base import NotifyBase

# Supported Dither Types
DOT_DITHER_TYPES = (
    "DIFFUSION",
    "ORDERED",
    "NONE",
)

# Supported Dither Kernels
DOT_DITHER_KERNELS = (
    "THRESHOLD",
    "ATKINSON",
    "BURKES",
    "FLOYD_STEINBERG",
    "SIERRA2",
    "STUCKI",
    "JARVIS_JUDICE_NINKE",
    "DIFFUSION_ROW",
    "DIFFUSION_COLUMN",
    "DIFFUSION_2D",
)

# Supported API modes; first entry is the default
DOT_MODES = ("text", "image")


class NotifyDot(NotifyBase):
    """A wrapper for Dot. Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Dot."

    # The services URL
    service_url = "https://dot.mindreset.tech"

    # All notification requests are secure
    secure_protocol = "dot"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://appriseit.com/services/dot/"

    # Allows the user to specify the NotifyImageSize object
    image_size = NotifyImageSize.XY_128

    # Support Attachments
    attachment_support = True

    # Define object templates
    templates = ("{schema}://{apikey}@{device_id}/",)

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "apikey": {
                "name": _("API Key"),
                "type": "string",
                "required": True,
                "private": True,
            },
            "device_id": {
                "name": _("Device Serial Number"),
                "type": "string",
                "required": True,
                "map_to": "device_id",
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "mode": {
                "name": _("API Mode"),
                "type": "choice:string",
                "values": DOT_MODES,
                "default": DOT_MODES[0],
            },
            "refresh": {
                "name": _("Refresh Now"),
                "type": "bool",
                "default": True,
                "map_to": "refresh_now",
            },
            "signature": {
                "name": _("Text Signature"),
                "type": "string",
            },
            "icon": {
                "name": _("Icon Base64 (Text API)"),
                "type": "string",
            },
            "image": {
                "name": _("Image Base64 (Image API)"),
                "type": "string",
                "map_to": "image_data",
            },
            "link": {
                "name": _("Link"),
                "type": "string",
            },
            "border": {
                "name": _("Border"),
                "type": "int",
                "min": 0,
                "max": 1,
                "default": 0,
            },
            "dither_type": {
                "name": _("Dither Type"),
                "type": "choice:string",
                "values": DOT_DITHER_TYPES,
                "default": "DIFFUSION",
            },
            "dither_kernel": {
                "name": _("Dither Kernel"),
                "type": "choice:string",
                "values": DOT_DITHER_KERNELS,
                "default": "FLOYD_STEINBERG",
            },
            "task_key": {
                "name": _("Task Key"),
                "type": "string",
            },
        },
    )

    def __init__(
        self,
        apikey=None,
        device_id=None,
        mode=DOT_MODES[0],
        refresh_now=None,
        signature=None,
        icon=None,
        link=None,
        border=None,
        dither_type=None,
        dither_kernel=None,
        image_data=None,
        task_key=None,
        **kwargs,
    ):
        """Initialize Notify Dot Object."""
        super().__init__(**kwargs)

        # API Key (from user)
        self.apikey = apikey

        # Device ID tracks the Dot hardware serial.
        self.device_id = device_id

        # Refresh Now flag: True shows content immediately (default).
        self.refresh_now = (
            parse_bool(
                refresh_now,
                self.template_args["refresh"]["default"],
            )
            if refresh_now is not None
            else self.template_args["refresh"]["default"]
        )

        # API mode
        self.mode = (
            mode.lower()
            if isinstance(mode, str) and mode.lower() in DOT_MODES
            else DOT_MODES[0]
        )
        if not isinstance(mode, str) or mode.lower() not in DOT_MODES:
            self.logger.warning(
                "Unsupported Dot mode (%s) specified; defaulting to '%s'.",
                mode,
                self.mode,
            )

        # Signature text used by the Text API footer.
        self.signature = signature if isinstance(signature, str) else None

        # Icon for the Text API (base64 PNG 40x40, lower-left corner).
        self.icon = icon if isinstance(icon, str) else None

        # Image payload for the Image API (base64 PNG 296x152).
        self.image_data = (
            image_data if image_data and isinstance(image_data, str) else None
        )

        # Link for tap-to-interact navigation.
        self.link = link if isinstance(link, str) else None

        # Border for the Image API
        self.border = border

        # Dither type for Image API
        self.dither_type = dither_type

        # Dither kernel for the Image API
        self.dither_kernel = dither_kernel

        # Task Key for specifying which content to update
        self.task_key = (
            task_key if task_key and isinstance(task_key, str) else None
        )

        # Text API endpoint (v2)
        self.text_api_url = (
            "https://dot.mindreset.tech/api/authV2/open/device/"
            f"{self.device_id}/text"
        )

        # Image API endpoint (v2)
        self.image_api_url = (
            "https://dot.mindreset.tech/api/authV2/open/device/"
            f"{self.device_id}/image"
        )

        return

    def _post(self, api_url, payload, headers):
        """POST payload to api_url; return True on success, False otherwise."""

        if self.logger.isEnabledFor(logging.DEBUG):
            self.logger.debug(
                "Dot POST URL: %s (cert_verify=%r)",
                api_url,
                self.verify_certificate,
            )
            self.logger.debug("Dot Payload %s", sanitize_payload(payload))

        # Always call throttle before any remote server i/o is made
        self.throttle()

        try:
            r = requests.post(
                api_url,
                data=json.dumps(payload),
                headers=headers,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )

            if r.status_code == requests.codes.ok:
                self.logger.info(
                    "Sent Dot notification to %s.", self.device_id
                )
                return True

            status_str = NotifyDot.http_response_code_lookup(r.status_code)
            self.logger.warning(
                "Failed to send Dot notification to %s: %s%serror=%d.",
                self.device_id,
                status_str,
                ", " if status_str else "",
                r.status_code,
            )
            self.logger.debug(
                "Response Details:\r\n%r", (r.content or b"")[:2000]
            )
            return False

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred sending Dot notification to %s.",
                self.device_id,
            )
            self.logger.debug("Socket Exception: %s", str(e))
            return False

    def _send_text(self, body, title, icon_data, headers):
        """Send body/title via the Text API."""
        payload = {"refreshNow": self.refresh_now}

        if title:
            payload["title"] = title
        if body:
            payload["message"] = body
        if self.signature:
            payload["signature"] = self.signature
        if icon_data:
            payload["icon"] = icon_data
        if self.link:
            payload["link"] = self.link
        if self.task_key is not None:
            payload["taskKey"] = self.task_key

        return self._post(self.text_api_url, payload, headers)

    def _send_image(self, image_data, headers):
        """Send image_data via the Image API."""
        payload = {
            "image": image_data,
            "refreshNow": self.refresh_now,
        }

        if self.link:
            payload["link"] = self.link
        if self.border is not None:
            payload["border"] = self.border
        if self.dither_type is not None:
            payload["ditherType"] = self.dither_type
        if self.dither_kernel is not None:
            payload["ditherKernel"] = self.dither_kernel
        if self.task_key is not None:
            payload["taskKey"] = self.task_key

        return self._post(self.image_api_url, payload, headers)

    def _resolve_attachment(self, attach, warn_label):
        """Return base64 string from the first usable attachment, or None."""
        if not (attach and self.attachment_support):
            return None
        if len(attach) > 1:
            self.logger.warning(
                "Multiple attachments provided; only the first"
                " one will be used as %s.",
                warn_label,
            )
        try:
            attachment = attach[0]
            if attachment:
                return attachment.base64()
        except Exception as e:
            self.logger.warning("Failed to process attachment: %s", str(e))
        return None

    def send(
        self,
        body,
        title="",
        notify_type=NotifyType.INFO,
        attach=None,
        **kwargs,
    ):
        """Perform Dot Notification."""

        if not self.apikey:
            self.logger.warning("No API key was specified")
            return False

        if not self.device_id:
            self.logger.warning("No device ID was specified")
            return False

        headers = {
            "Authorization": f"Bearer {self.apikey}",
            "Content-Type": "application/json",
            "User-Agent": self.app_id,
        }

        if self.mode == "image":
            # Image-only mode: body and title are ignored.
            if title or body:
                self.logger.warning(
                    "Title and body are not supported in image mode"
                    " and will be ignored."
                )
            image_data = (
                self.image_data
                if isinstance(self.image_data, str)
                else self._resolve_attachment(attach, "image")
            )
            if not image_data:
                self.logger.warning(
                    "Image mode selected but no image data was provided."
                )
                return False
            return self._send_image(image_data, headers)

        # Text mode (default): smart dual-send.
        # Body/title go to the Text API; image data or attachment goes
        # to the Image API.  Text is always dispatched before image.
        image_data = (
            self.image_data
            if isinstance(self.image_data, str)
            else self._resolve_attachment(attach, "image")
        )
        has_text = bool(body or title)
        has_image = bool(image_data)

        if not has_text and not has_image:
            self.logger.warning(
                "Nothing to send to Dot. device %s.", self.device_id
            )
            return False

        has_error = False
        if has_text and not self._send_text(body, title, self.icon, headers):
            has_error = True

        if has_image and not self._send_image(image_data, headers):
            has_error = True
        return not has_error

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another similar one.
        """
        return (
            self.secure_protocol,
            self.apikey,
            self.device_id,
            self.mode,
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        params = {}
        params["refresh"] = "yes" if self.refresh_now else "no"

        # Include mode only when non-default
        if self.mode != DOT_MODES[0]:
            params["mode"] = self.mode

        if self.signature:
            params["signature"] = self.signature
        if self.icon:
            params["icon"] = self.icon
        if self.image_data:
            params["image"] = self.image_data
        if self.border is not None:
            params["border"] = str(self.border)
        if self.dither_type and self.dither_type != "DIFFUSION":
            params["dither_type"] = self.dither_type
        if self.dither_kernel and self.dither_kernel != "FLOYD_STEINBERG":
            params["dither_kernel"] = self.dither_kernel
        if self.link:
            params["link"] = self.link
        if self.task_key:
            params["task_key"] = self.task_key

        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        return "{schema}://{apikey}@{device_id}/?{params}".format(
            schema=self.secure_protocol,
            apikey=self.pprint(
                self.apikey, privacy, mode=PrivacyMode.Secret, safe=""
            ),
            device_id=NotifyDot.quote(self.device_id, safe=""),
            params=NotifyDot.urlencode(params),
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        return 1 if self.device_id else 0

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to
        re-instantiate this object."""

        results = NotifyBase.parse_url(url)
        if not results:
            return results

        # Determine mode: explicit ?mode= wins; path provides backward compat.
        mode = DOT_MODES[0]
        if results["qsd"].get("mode"):
            candidate = results["qsd"]["mode"].lower().strip()
            if candidate in DOT_MODES:
                mode = candidate
            else:
                NotifyDot.logger.warning(
                    "Unsupported Dot mode (%s) specified; defaulting to '%s'.",
                    candidate,
                    DOT_MODES[0],
                )
        else:
            path_tokens = NotifyDot.split_path(results.get("fullpath"))
            if path_tokens:
                candidate = path_tokens[0].lower()
                if candidate in DOT_MODES:
                    mode = candidate
                else:
                    NotifyDot.logger.warning(
                        "Unsupported Dot mode (%s) specified;"
                        " defaulting to '%s'.",
                        candidate,
                        DOT_MODES[0],
                    )
        results["mode"] = mode

        # Extract API key from user
        if results.get("user"):
            results["apikey"] = NotifyDot.unquote(results["user"])

        # Extract device ID from hostname
        if results.get("host"):
            results["device_id"] = NotifyDot.unquote(results["host"])

        # Refresh Now
        refresh_value = results["qsd"].get("refresh")
        if refresh_value:
            results["refresh_now"] = parse_bool(refresh_value.strip())

        # Signature
        if results["qsd"].get("signature"):
            results["signature"] = NotifyDot.unquote(
                results["qsd"]["signature"].strip()
            )

        # Icon
        if results["qsd"].get("icon"):
            results["icon"] = NotifyDot.unquote(results["qsd"]["icon"].strip())

        # Link
        if results["qsd"].get("link"):
            results["link"] = NotifyDot.unquote(results["qsd"]["link"].strip())

        # Border
        if results["qsd"].get("border"):
            with suppress(TypeError, ValueError):
                results["border"] = int(results["qsd"]["border"].strip())

        # Dither Type
        if results["qsd"].get("dither_type"):
            results["dither_type"] = NotifyDot.unquote(
                results["qsd"]["dither_type"].strip()
            )

        # Dither Kernel
        if results["qsd"].get("dither_kernel"):
            results["dither_kernel"] = NotifyDot.unquote(
                results["qsd"]["dither_kernel"].strip()
            )

        # Image (Image API)
        if results["qsd"].get("image"):
            results["image_data"] = NotifyDot.unquote(
                results["qsd"]["image"].strip()
            )

        # Task Key
        if results["qsd"].get("task_key"):
            results["task_key"] = NotifyDot.unquote(
                results["qsd"]["task_key"].strip()
            )

        return results



#
# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# PushPlus is a Chinese notification platform that delivers messages via
# WeChat and several other channels.  You can find its API documentation at:
#   https://www.pushplus.plus/doc/guide/api.html
#
# To obtain your personal token:
#   1. Register or sign in at https://www.pushplus.plus/
#   2. Copy the token shown on the dashboard under the "Push" section.
#
# Group (topic) sending is also supported.  After creating a group under
# "Group Push" in the PushPlus console, use the group code as the topic:
#   https://www.pushplus.plus/doc/guide/group.html
#
# Basic Apprise URL forms:
#
#   Personal notification (WeChat default):
#     pushplus://{token}
#
#   Group topic -- one notification per topic:
#     pushplus://{token}/{topic}
#     pushplus://{token}/{topic1}/{topic2}
#
#   Select a delivery channel (mode= and channel= are synonyms):
#     pushplus://{token}?channel=mail
#     pushplus://{token}?mode=cp
#
#   Topic + delivery channel:
#     pushplus://{token}/{topic}?channel=mail
#
#   Webhook channel with a named endpoint -- two equivalent forms:
#     pushplus://{token}?channel=webhook&name={webhook_name}
#     pushplus://{webhook_name}@{token}
#
#   When the schema://{name}@{token} form is used and no explicit channel=
#   is given, the webhook channel is implied automatically.
#
#   Native PushPlus API URL (also accepted by parse_native_url):
#     https://www.pushplus.plus/send?token={token}
#
# Schema alias for WeCom users:
#   wecom://{token}    -- identical to pushplus://{token}?channel=cp
#
# Note: wechat:// is reserved for a future direct WeChat Official Account
# API plugin and is not supported here.
#
# API References:
#   https://www.pushplus.plus/doc/guide/api.html
#   https://www.pushplus.plus/doc/guide/group.html

import json
import re

import requests

from ..common import NotifyFormat, NotifyType
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from ..utils.parse import parse_list, validate_regex
from .base import NotifyBase

# PushPlus application-level response codes.
# The HTTP status is always 200; the real result lives in the JSON body.
# Reference: https://www.pushplus.plus/doc/guide/api.html
PUSHPLUS_RESPONSE_CODES = {
    200: "Request succeeded.",
    900: "System exception.",
    903: "Sending failed.",
    905: "Request parameter error.",
    907: "Token does not exist.",
    908: "User is blocked.",
    909: "Content requires review before sending.",
    912: "No available service package.",
}

# Map Apprise's standard notify_format values to the PushPlus template
# identifiers.  PushPlus uses these to render the body server-side before
# delivery -- the content itself does not change; only the rendering hint.
PUSHPLUS_FORMAT_MAP = {
    NotifyFormat.HTML: "html",
    NotifyFormat.MARKDOWN: "markdown",
    NotifyFormat.TEXT: "txt",
}

# Default PushPlus template when the format has no explicit mapping
PUSHPLUS_FORMAT_DEFAULT = "html"


class PushPlusChannel:
    """Defines the PushPlus delivery channels.

    The channel controls where the rendered notification is delivered.
    It is supplied as the channel= (or mode=) query parameter in the
    Apprise URL, e.g. pushplus://{token}?channel=mail.
    """

    # Deliver via WeChat (PushPlus default)
    WECHAT = "wechat"

    # Deliver via a configured webhook endpoint
    WEBHOOK = "webhook"

    # Deliver via WeCom (WeChat Work / Enterprise WeChat).
    # The API value "cp" is PushPlus's internal identifier for this channel.
    # Both "cp" and the friendly alias "wecom" are accepted; both resolve here.
    WECOM = "cp"

    # Deliver via email
    MAIL = "mail"

    # Deliver via SMS
    SMS = "sms"


# All valid PushPlus delivery channels (these are the API values)
PUSHPLUS_CHANNELS = (
    PushPlusChannel.WECHAT,
    PushPlusChannel.WEBHOOK,
    PushPlusChannel.WECOM,
    PushPlusChannel.MAIL,
    PushPlusChannel.SMS,
)

# The PushPlus default delivery channel (WeChat)
PUSHPLUS_CHANNEL_DEFAULT = PushPlusChannel.WECHAT

# A group topic has no special prefix; alphanumeric codes up to 50 chars.
IS_TOPIC = re.compile(
    r"^(?P<topic>[a-z0-9_-]{1,50})$",
    re.I,
)

# Schema names that auto-select a specific delivery channel when used
# in place of the default "pushplus" schema.  Mirrors the Kodi/XBMC pattern.
# Note: wechat:// is intentionally omitted -- it is reserved for a future
# direct WeChat Official Account API integration.
PUSHPLUS_SCHEMA_MAP = {
    # wecom:// forces channel=cp (WeCom / Enterprise WeChat)
    "wecom": PushPlusChannel.WECOM,
}


class NotifyPushplus(NotifyBase):
    """A wrapper for PushPlus Notifications."""

    # The default descriptive name associated with the Notification
    service_name = _("Pushplus")

    # The services URL
    service_url = "https://www.pushplus.plus/"

    # The default protocol; wecom:// is accepted as a schema alias.
    # Note: wechat:// is reserved for a future direct WeChat Official
    # Account API plugin and is intentionally not listed here.
    secure_protocol = ("pushplus", "wecom")

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://appriseit.com/services/pushplus/"

    # URL used to POST notifications
    notify_url = "https://www.pushplus.plus/send"

    # Maximum body length documented by PushPlus
    body_maxlen = 20000

    # Title is capped at a safe limit (not explicitly documented by PushPlus)
    title_maxlen = 200

    # Default to HTML since PushPlus renders HTML by default
    notify_format = NotifyFormat.HTML

    # Define object URL templates
    templates = (
        # No topics: personal notification with optional ?channel= override
        "{schema}://{token}",
        # Topics in path: one API call per topic
        "{schema}://{token}/{targets}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "token": {
                "name": _("User Token"),
                "type": "string",
                "private": True,
                "required": True,
                # PushPlus tokens are 32-64 alphanumeric/underscore/dash chars
                "regex": (r"^[a-z0-9_-]{32,64}$", "i"),
            },
            # Group topics go directly in the URL path with no prefix
            "targets": {
                "name": _("Group Topics"),
                "type": "list:string",
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            # Allow the token to be supplied as a query parameter
            "token": {
                "alias_of": "token",
            },
            # ?to= is the standard Apprise alias for targets (topics)
            "to": {
                "alias_of": "targets",
            },
            # Delivery channel -- selects where the message is delivered.
            # One of the PUSHPLUS_CHANNELS values (wechat is the default).
            "channel": {
                "name": _("Channel"),
                "type": "choice:string",
                "values": PUSHPLUS_CHANNELS,
                "default": PUSHPLUS_CHANNEL_DEFAULT,
            },
            # mode= is an alias for channel=; the two are fully synonymous.
            "mode": {
                "alias_of": "channel",
            },
            # ?topic= backward-compat alias that maps to targets
            "topic": {
                "alias_of": "targets",
            },
            # Webhook endpoint name; only meaningful when channel=webhook.
            # The URL parameter is ?name= but the __init__ kwarg is webhook=
            # to avoid shadowing URLBase.name.
            "name": {
                "name": _("Webhook Name"),
                "type": "string",
                "map_to": "webhook",
            },
        },
    )

    def __init__(
        self, token, targets=None, channel=None, webhook=None, **kwargs
    ):
        """Initialize Pushplus Object."""
        super().__init__(**kwargs)

        # Validate the required user token
        self.token = validate_regex(
            token,
            *self.template_tokens["token"]["regex"],
        )
        if not self.token:
            msg = "The Pushplus token ({}) is invalid.".format(token)
            self.logger.warning(msg)
            raise TypeError(msg)

        # Resolve the delivery channel from the schema alias when the URL
        # was entered as wechat:// or wecom:// instead of pushplus://.
        # self.schema is set by NotifyBase from the parsed URL protocol.
        schema_channel = PUSHPLUS_SCHEMA_MAP.get((self.schema or "").lower())

        # Determine the active delivery channel:
        #   1. Explicit channel= / mode= argument (already resolved by
        #      the caller since mode is an alias_of channel in template_args)
        #   2. Schema-implied channel (wechat:// or wecom://)
        #   3. Default (WeChat)
        if channel:
            # Normalise the 'wecom' friendly alias to the API value 'cp'
            if channel.lower() == "wecom":
                channel = PushPlusChannel.WECOM
            self.channel = next(
                (c for c in PUSHPLUS_CHANNELS if c == channel.lower()),
                None,
            )
            if not self.channel:
                msg = "The Pushplus channel ({}) is not valid.".format(channel)
                self.logger.warning(msg)
                raise TypeError(msg)

        elif schema_channel:
            # Schema-implied channel (wechat:// or wecom://)
            self.channel = schema_channel

        else:
            # Default to WeChat
            self.channel = PUSHPLUS_CHANNEL_DEFAULT

        # Resolved group topics -- one API call is made per topic
        self.topics = []

        # Preserve unrecognised targets for round-trip fidelity in url()
        self.invalid_targets = []

        # Parse each target from the list -- only plain topics are valid here
        for target in parse_list(targets):
            result = IS_TOPIC.match(target)
            if result:
                self.topics.append(result.group("topic"))
                continue

            # Unrecognised entry -- log and preserve
            self.logger.warning("Dropped invalid Pushplus topic: %s", target)
            self.invalid_targets.append(target)

        # Store the webhook name; only meaningful when channel=webhook.
        # Kept as self.webhook (not self.name) to avoid shadowing URLBase.name.
        # Arrives via the ?name= URL parameter (mapped to webhook= by map_to).
        self.webhook = (
            webhook if isinstance(webhook, str) and webhook else None
        )

        return

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform PushPlus Notification."""

        # Prepare our headers
        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json",
        }

        # Derive the PushPlus rendering template from Apprise's standard
        # notify_format.  This is a server-side rendering hint only; the
        # content payload does not change.
        pp_template = PUSHPLUS_FORMAT_MAP.get(
            self.notify_format, PUSHPLUS_FORMAT_DEFAULT
        )

        # When no topics are configured, fall back to a single personal send
        # by iterating over a list containing None as the sole entry.
        topics_to_notify = self.topics if self.topics else [None]

        # Track whether any individual send failed
        has_error = False

        for topic in topics_to_notify:
            # Build the payload for this particular topic
            payload = {
                # Authentication token
                "token": self.token,
                # Fall back to the body when no title is provided
                "title": title if title else body,
                # Notification body content
                "content": body,
                # Rendering template derived from notify_format
                "template": pp_template,
                # Delivery channel
                "channel": self.channel,
            }

            # Add the group topic when sending to a specific group
            if topic:
                payload["topic"] = topic

            # Add the webhook name when the webhook channel is selected
            if self.channel == PushPlusChannel.WEBHOOK and self.webhook:
                payload["webhook"] = self.webhook

            # Debug logging so the caller can inspect what will be sent
            self.logger.debug(
                "PushPlus POST URL: %s (cert_verify=%r)",
                self.notify_url,
                self.verify_certificate,
            )
            self.logger.debug("PushPlus Payload: %r", payload)

            # Always throttle before each remote server I/O call
            self.throttle()

            try:
                r = requests.post(
                    self.notify_url,
                    headers=headers,
                    # Encode explicitly for non-ASCII (e.g. Chinese) chars
                    data=json.dumps(payload, ensure_ascii=False).encode(
                        "utf-8"
                    ),
                    verify=self.verify_certificate,
                    timeout=self.request_timeout,
                )

                if r.status_code != requests.codes.ok:
                    # HTTP-level failure -- log it and move on
                    status_str = NotifyPushplus.http_response_code_lookup(
                        r.status_code
                    )
                    self.logger.warning(
                        "Failed to send PushPlus notification: "
                        "{}{}error={}.".format(
                            status_str,
                            ", " if status_str else "",
                            r.status_code,
                        )
                    )
                    self.logger.debug(
                        "Response Details:\r\n%r",
                        (r.content or b"")[:2000],
                    )
                    # Mark our failure and continue with the next topic
                    has_error = True
                    continue

                # PushPlus always returns HTTP 200; the real result is in
                # the JSON body where code == 200 means success.
                try:
                    content = json.loads(r.content)
                except (AttributeError, TypeError, ValueError):
                    # ValueError = r.content is unparsable
                    # TypeError  = r.content is None
                    # AttributeError = r is None
                    content = {}

                # Check the application-level status code
                api_code = content.get("code") if content else None
                if api_code != 200:
                    # Application-level failure
                    error_str = PUSHPLUS_RESPONSE_CODES.get(
                        api_code,
                        # Fall back to the msg field, then a generic string
                        (
                            content.get("msg", "Unknown error")
                            if content
                            else "Unknown error"
                        ),
                    )
                    self.logger.warning(
                        "Failed to send PushPlus notification: "
                        "code={}: {}.".format(api_code, error_str)
                    )
                    self.logger.debug(
                        "Response Details:\r\n%r",
                        content if content else (r.content or b"")[:2000],
                    )
                    # Mark our failure and continue with the next topic
                    has_error = True
                    continue

            except requests.RequestException as e:
                self.logger.warning(
                    "A Connection error occurred sending"
                    " PushPlus notification."
                )
                self.logger.debug("Socket Exception: %s", str(e))
                # Mark our failure and continue with the next topic
                has_error = True
                continue

            # Notification delivered for this topic
            self.logger.info(
                "Sent PushPlus notification%s.",
                " to topic {}".format(topic) if topic else "",
            )

        return not has_error

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another similar one.

        Targets or end points should never be identified here.
        """
        # The token is the sole connection identity for PushPlus
        return (self.secure_protocol[0], self.token)

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # When channel=webhook with a named endpoint, use the compact
        # {name}@pushplus://{token} form.  Both ?channel=webhook and ?name=
        # are implied by the user@ prefix and omitted from the query string.
        webhook_prefix = (
            self.channel == PushPlusChannel.WEBHOOK and self.webhook
        )

        # Start with an empty params dict
        params = {}

        # When not using the webhook prefix form, include ?channel= when it
        # differs from the default.  The schema alias (wechat:// / wecom://)
        # is never emitted here -- we always normalise back to pushplus://
        # plus ?channel= for clarity.
        if not webhook_prefix and self.channel != PUSHPLUS_CHANNEL_DEFAULT:
            params["channel"] = self.channel

        # Merge in standard Apprise URL parameters (verify, format, etc.)
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        # Build the targets path from group topics and any invalid entries
        targets = list(self.topics) + list(self.invalid_targets)

        # Masked token for the URL
        token_str = self.pprint(
            self.token,
            privacy,
            mode=PrivacyMode.Secret,
            safe="",
        )

        # When using the webhook prefix form, the endpoint name goes in the
        # user@ position (schema://name@token) -- channel=webhook and ?name=
        # are both implied by the user@ presence and omitted from params.
        if webhook_prefix:
            name_str = NotifyPushplus.quote(self.webhook, safe="")
            if targets:
                return "{schema}://{name}@{token}/{targets}/?{params}".format(
                    schema=self.secure_protocol[0],
                    name=name_str,
                    token=token_str,
                    targets="/".join(
                        NotifyPushplus.quote(t, safe="") for t in targets
                    ),
                    params=NotifyPushplus.urlencode(params),
                )
            return "{schema}://{name}@{token}/?{params}".format(
                schema=self.secure_protocol[0],
                name=name_str,
                token=token_str,
                params=NotifyPushplus.urlencode(params),
            )

        if targets:
            # One or more topics: include them in the URL path
            return "{schema}://{token}/{targets}/?{params}".format(
                schema=self.secure_protocol[0],
                token=token_str,
                targets="/".join(
                    NotifyPushplus.quote(t, safe="") for t in targets
                ),
                params=NotifyPushplus.urlencode(params),
            )

        # No topics -- simple personal notification URL
        return "{schema}://{token}/?{params}".format(
            schema=self.secure_protocol[0],
            token=token_str,
            params=NotifyPushplus.urlencode(params),
        )

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to
        re-instantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Prefer ?token= query parameter over the URL host field
        if "token" in results["qsd"] and results["qsd"]["token"]:
            # Token was supplied as a query parameter
            results["token"] = NotifyPushplus.unquote(results["qsd"]["token"])
        else:
            # Token is the URL host
            results["token"] = NotifyPushplus.unquote(results["host"])

        # Collect group topics from the URL path
        results["targets"] = NotifyPushplus.split_path(results["fullpath"])

        # ?to= appends additional targets (comma/space delimited supported)
        if "to" in results["qsd"] and results["qsd"]["to"]:
            results["targets"] += NotifyPushplus.parse_list(
                results["qsd"]["to"]
            )

        # ?topic= backward-compat alias also appends topics
        if "topic" in results["qsd"] and results["qsd"]["topic"]:
            results["targets"] += NotifyPushplus.parse_list(
                results["qsd"]["topic"]
            )

        # Extract the delivery channel from ?channel= or its alias ?mode=
        # mode= takes lower priority; channel= wins if both are present.
        if "mode" in results["qsd"] and results["qsd"]["mode"]:
            results["channel"] = NotifyPushplus.unquote(results["qsd"]["mode"])
        if "channel" in results["qsd"] and results["qsd"]["channel"]:
            results["channel"] = NotifyPushplus.unquote(
                results["qsd"]["channel"]
            )

        # Extract the webhook name -- users specify it as ?name= in the URL.
        # We store it internally as 'webhook' to avoid shadowing URLBase.name.
        if "name" in results["qsd"] and results["qsd"]["name"]:
            results["webhook"] = NotifyPushplus.unquote(results["qsd"]["name"])

        # Support the {webhook_name}@pushplus://{token} short form.
        # When user@ is present, it identifies the webhook endpoint name.
        # If no channel was explicitly given, webhook is the implied channel.
        if results.get("user"):
            # Use user@ as the webhook name when ?name= was not also supplied
            if "webhook" not in results:
                results["webhook"] = NotifyPushplus.unquote(results["user"])
            # Imply webhook channel when no explicit channel was specified
            if "channel" not in results:
                results["channel"] = PushPlusChannel.WEBHOOK

        return results

    @staticmethod
    def parse_native_url(url):
        """Support native PushPlus API URLs of the form:
        https://www.pushplus.plus/send?token=TOKEN[&other_params]
        """
        result = re.match(
            r"^https?://www\.pushplus\.plus/send"
            r"(?:\?(?P<params>[^#]+))?$",
            url,
            re.I,
        )
        if result:
            params = result.group("params") or ""
            # The token must be present as a query parameter
            tok = re.search(
                r"(?:(?:^|&))token=(?P<token>[a-z0-9_-]+)",
                params,
                re.I,
            )
            if tok:
                # Re-build as an Apprise URL.  Preserve all existing params
                # so that topic, channel, and name all round-trip correctly.
                return NotifyPushplus.parse_url(
                    "{schema}://{token}/?{params}".format(
                        schema=NotifyPushplus.secure_protocol[0],
                        token=tok.group("token"),
                        params=params,
                    )
                )
        return None



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""XMPP Notifications"""

from __future__ import annotations

import re
from typing import Any, Optional

from ...common import NotifyType
from ...locale import gettext_lazy as _
from ...url import PrivacyMode
from ...utils.parse import parse_bool, parse_list, validate_regex
from ..base import NotifyBase
from .adapter import SLIXMPP_SUPPORT_AVAILABLE, SlixmppAdapter, XMPPConfig
from .common import SECURE_MODES, SecureXMPPMode

# A pragmatic, "hardened" JID validator intended for Apprise URLs.
#
# - Supports: local@domain and local@domain/resource
# - Rejects whitespace anywhere
# - Rejects missing local or domain
# - Rejects '@' in the domain component
#
# This does not try to fully implement RFC 7622. The goal is to catch bad
# inputs early and reliably while still supporting common JID patterns.
IS_JID = re.compile(
    r"^\s*(?P<is_room>#|%23)?(?P<local>[^@\s/]+)((@|%40)"
    r"(?P<domain>[^@\s/]+))?(?:(/|%2F)(?P<resource>[^%/\s]+)"
    r"((/|%2F).*)?)?\s*$"
)


class NotifyXMPP(NotifyBase):
    """Send notifications via XMPP using Slixmpp."""

    # Set our global enabled flag
    enabled = SLIXMPP_SUPPORT_AVAILABLE and SlixmppAdapter._enabled

    requirements = {
        # Define our required packaging in order to work
        "packages_required": SlixmppAdapter.package_dependency(),
    }

    # The default descriptive name associated with the Notification
    service_name = "XMPP"

    # The services URL
    service_url = "https://xmpp.org/"

    # The default insecure protocol
    protocol = "xmpp"

    # The default secure protocol
    secure_protocol = "xmpps"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://appriseit.com/services/xmpp/"

    templates = (
        "{schema}://{user}:{password}@{host}",
        "{schema}://{user}:{password}@{host}:{port}",
        "{schema}://{user}:{password}@{host}/{targets}",
        "{schema}://{user}:{password}@{host}:{port}/{targets}",
    )

    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "host": {
                "name": _("Hostname"),
                "type": "string",
                "required": True,
            },
            "port": {
                "name": _("Port"),
                "type": "int",
                "min": 1,
                "max": 65535,
            },
            "user": {
                "name": _("User"),
                "type": "string",
                "required": True,
            },
            "password": {
                "name": _("Password"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "target_user": {
                "name": _("Target User"),
                "type": "string",
                "map_to": "targets",
            },
            "target_channels": {
                "name": _("Target Channel"),
                "type": "string",
                "prefix": "#",
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
            },
        },
    )

    template_args = dict(
        NotifyBase.template_args,
        **{
            "xmpp": {
                "name": _("XMPP Server"),
                "type": "string",
                "map_to": "xmpp_host",
            },
            "mode": {
                "name": _("Secure Mode"),
                "type": "choice:string",
                "values": SECURE_MODES,
                "default": SecureXMPPMode.STARTTLS,
                "map_to": "secure_mode",
            },
            "roster": {
                "name": _("Get Roster"),
                "type": "bool",
                "default": False,
            },
            "subject": {
                "name": _("Use Subject"),
                "type": "bool",
                "default": False,
            },
            "keepalive": {
                "name": _("Keep Connection Alive"),
                "type": "bool",
                "default": False,
            },
            "to": {"alias_of": "targets"},
            "name": {
                "name": _("MUC Nickname"),
                "type": "string",
            },
        },
    )

    def __init__(
        self,
        targets: Optional[list[(str, str)]] = None,
        secure_mode: Optional[str] = None,
        roster: Optional[bool] = None,
        subject: Optional[bool] = None,
        keepalive: Optional[bool] = None,
        name: Optional[str] = None,
        xmpp_host: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)

        # xmpp_host allows the connection host to differ from the JID domain.
        # Mirrors the smtp= / smtp_host pattern in the email plugin.
        self.xmpp_host = (
            xmpp_host.strip()
            if isinstance(xmpp_host, str) and xmpp_host.strip()
            else ""
        )

        try:
            self.jid, _ = self.normalize_jid(self.user or "", self.host)

        except ValueError:
            msg = f"An invalid XMPP JID ({self.user}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg) from None

        self.targets: list[(str, str)] = []
        # Flag for tracking if we want Multi-User Chat function enabled
        self.want_muc = False

        for target in parse_list(targets):
            try:
                jid, is_muc = self.normalize_jid(target or "", self.host)
                mtype = "groupchat" if is_muc else "chat"
                if is_muc:
                    self.want_muc = True

            except ValueError:
                self.logger.warning(
                    "Dropped invalid XMPP target (%s).", target
                )
                continue
            self.targets.append((mtype, jid))

        if isinstance(secure_mode, str) and secure_mode.strip():
            self.secure_mode = secure_mode.strip().lower()
            self.secure_mode = next(
                (k for k in SECURE_MODES if k.startswith(self.secure_mode)),
                None,
            )
            if self.secure_mode not in SECURE_MODES:
                msg = (
                    "The XMPP secure mode specified "
                    f"({secure_mode}) is invalid."
                )
                self.logger.warning(msg)
                raise TypeError(msg)

        else:
            self.secure_mode = (
                SecureXMPPMode.NONE
                if not self.secure
                else self.template_args["mode"]["default"]
            )

        # Prepare our roster check
        self.roster = (
            self.template_args["roster"]["default"]
            if roster is None
            else bool(roster)
        )

        self.subject = (
            self.template_args["subject"]["default"]
            if subject is None
            else bool(subject)
        )

        self.keepalive = (
            self.template_args["keepalive"]["default"]
            if keepalive is None
            else bool(keepalive)
        )

        if self.secure and self.secure_mode == SecureXMPPMode.NONE:
            self.secure_mode = self.template_args["mode"]["default"]
            self.logger.warning(
                "Ambiguous XMPP configuration: secure=True and mode=None; "
                "secure setting prevails; setting mode=%s",
                self.secure_mode,
            )

        elif not self.secure and self.secure_mode != SecureXMPPMode.NONE:
            self.logger.warning(
                "Ambiguous XMPP configuration: secure=False and mode=%s; "
                "mode setting prevails; setting secure=True",
                self.secure_mode,
            )
            self.secure = True

        # MUC nickname: alphanumeric + underscore; falls back to the JID
        # username, then the app_id as a last resort
        self.name = validate_regex(name, r"^[a-zA-Z0-9_]+$") if name else None
        if self.name is None:
            self.name = self.user or self.app_id

        # Keepalive adapter (created lazily)
        self._adapter: Optional[SlixmppAdapter] = None

    def __del__(self) -> None:
        """Best-effort close for keepalive sessions."""
        try:
            if self._adapter is not None:
                self._adapter.close()

        except Exception:
            # Never raise from __del__
            pass

    @property
    def url_identifier(self) -> tuple[str, str, str, str, Optional[int]]:
        """Return the pieces that uniquely identify this configuration."""
        return (
            self.secure_protocol if self.secure else self.protocol,
            self.host,
            self.xmpp_host,
            self.user,
            self.password,
            self.port,
        )

    def url(self, privacy: bool = False, *args: Any, **kwargs: Any) -> str:
        """Return the URL representation of this notification."""

        # Initialize our parameters
        params = {
            "mode": self.secure_mode,
            "roster": "yes" if self.roster else "no",
            "subject": "yes" if self.subject else "no",
            "keepalive": "yes" if self.keepalive else "no",
        }

        # Only include name when it differs from the default
        # (JID user / app_id)
        if self.name != (self.user or self.app_id):
            params["name"] = self.name

        if self.xmpp_host and self.xmpp_host != self.host:
            params["xmpp"] = self.xmpp_host

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        auth = "{user}:{password}@".format(
            user=self.quote(self.jid, safe=""),
            password=self.pprint(
                self.password,
                privacy,
                mode=PrivacyMode.Secret,
                safe="",
            ),
        )

        default_port = SECURE_MODES[self.secure_mode]["default_port"]
        port = self.port if isinstance(self.port, int) else default_port
        port_str = "" if port == default_port else f":{port}"

        schema = self.secure_protocol if self.secure else self.protocol

        # Targets can contain '/' as a resource separator, so ensure it is
        # always percent-encoded in the path (otherwise Apprise will split it).
        # Use %23 for the MUC '#' prefix so it is not misread as a fragment.
        targets = "/".join(
            ("%23" if mode == "groupchat" else "") + self.quote(jid, safe="")
            for (mode, jid) in self.targets
        )

        return "{schema}://{auth}{host}{port}/{targets}?{params}".format(
            schema=schema,
            auth=auth,
            host=self.host,
            port=port_str,
            targets=targets,
            params=self.urlencode(params),
        )

    def send(
        self,
        body: str,
        title: str = "",
        notify_type: NotifyType = NotifyType.INFO,
        **kwargs: Any,
    ) -> bool:
        """Send a notification to one or more XMPP targets."""

        default_port = SECURE_MODES[self.secure_mode]["default_port"]

        self.throttle()

        config = XMPPConfig(
            jid=self.jid,
            password=self.password or "",
            host=self.xmpp_host or self.host,
            port=self.port if self.port else default_port,
            secure=self.secure_mode,
            verify_certificate=self.verify_certificate,
        )

        self.logger.debug(
            "XMPP init: jid=%s host=%s port=%d mode=%s "
            "verify_certificate=%s subject=%s roster=%s keepalive=%s "
            "targets=%s",
            self.jid,
            config.host,
            config.port,
            config.secure,
            config.verify_certificate,
            "yes" if self.subject else "no",
            "yes" if self.roster else "no",
            "yes" if self.keepalive else "no",
            self.targets,
        )

        subject = title if self.subject else ""

        if self.keepalive and self._adapter:
            # Reuse existing adapter
            return self._adapter.send_message(
                targets=self.targets,
                subject=subject,
                body=body,
            )

        adapter_kwargs = {
            "config": config,
            "targets": self.targets,
            "subject": subject,
            "body": body,
            "timeout": self.socket_connect_timeout,
            "roster": self.roster,
            "keepalive": self.keepalive,
            "want_muc": self.want_muc,
            "default_nickname": self.name,
        }
        if not self.keepalive:
            # One-shot mode: Create, process, and discard
            return SlixmppAdapter(**adapter_kwargs).process()

        # Keepalive mode, reuse a single adapter instance
        self._adapter = SlixmppAdapter(**adapter_kwargs)
        return self._adapter.send_message()

    @property
    def title_maxlen(self) -> Optional[int]:
        """
        Depending on if the subject field is set, we can control
        how the message is constructed.
        """

        return 0 if not self.subject else super().title_maxlen

    @staticmethod
    def normalize_jid(value: str, default_host: str) -> tuple[str, bool]:
        """Normalize and validate a JID.

        Behaviour:
        - If value is 'user' then it becomes 'user@default_host'.
        - If value is 'user@host' then it becomes 'user@host'.
        - If value is 'user@host/resource' then it becomes
           'user@host/resource'.
        - If value is 'user/resource' then it becomes
           'user@default_host/resource'.
        - If value already contains '@', it is used as-is, including an
           optional '/resource' suffix.
        """
        raw = (value or "").strip()
        results = IS_JID.match(raw)
        if not results:
            raise ValueError("Invalid JID")

        is_muc = bool(results.group("is_room"))
        host = results.group("domain") or default_host

        jid = f"{results.group('local')}@{host}"
        if results.group("resource"):
            jid = f"{jid}/{results.group('resource')}"

        return jid, is_muc

    @staticmethod
    def parse_url(url: str) -> Optional[dict[str, Any]]:
        """Parse an XMPP URL into constructor arguments."""
        results = NotifyBase.parse_url(url)
        if not results:
            return None

        # Targets from path
        results["targets"] = [
            NotifyXMPP.unquote(t)
            for t in NotifyXMPP.split_path(results.get("fullpath"))
        ]

        qd = results.get("qsd", {})

        # Support to= alias
        if "to" in qd and qd.get("to"):
            results["targets"] += NotifyXMPP.parse_list(
                NotifyXMPP.unquote(qd.get("to"))
            )

        if "mode" in results["qsd"] and len(results["qsd"]["mode"]):
            # Extract the secure mode to over-ride the default
            results["secure_mode"] = results["qsd"]["mode"].lower()

        if "roster" in results["qsd"] and len(results["qsd"]["roster"]):
            results["roster"] = parse_bool(results["qsd"]["roster"])

        if "subject" in results["qsd"] and len(results["qsd"]["subject"]):
            results["subject"] = parse_bool(results["qsd"]["subject"])

        if "keepalive" in results["qsd"] and len(results["qsd"]["keepalive"]):
            results["keepalive"] = parse_bool(results["qsd"]["keepalive"])

        if "name" in results["qsd"] and len(results["qsd"]["name"]):
            results["name"] = NotifyXMPP.unquote(results["qsd"]["name"])

        if "xmpp" in results["qsd"] and len(results["qsd"]["xmpp"]):
            results["xmpp_host"] = NotifyXMPP.unquote(results["qsd"]["xmpp"])

        return results

    @staticmethod
    def runtime_deps():
        """Return a tuple of top-level Python package names that this plugin
        imported as optional runtime dependencies.
        """
        return ("slixmpp",)
